#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the dataset object permission."""

import base64
import hmac
import mimetypes
from datetime import datetime, timezone
from hashlib import sha1, sha256
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional
from xml.etree import ElementTree

from graviti.exception import ResponseError
from graviti.openapi import do, get_object_permission
from graviti.utility import UserResponse, config, convert_datetime_to_gmt

if TYPE_CHECKING:
    from graviti.manager import Dataset

_EXPIRED_IN_SECOND = 600


class ObjectPermissionManager:
    """The basic structure of the object permission of the dataset.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.

    """

    _get_permission: Optional[Dict[str, Any]] = None
    _put_permission: Optional[Dict[str, Any]] = None

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _request_permission(self, actions: str) -> Dict[str, str]:
        dataset = self._dataset
        return get_object_permission(  # type: ignore[no-any-return]
            dataset.access_key,
            dataset.url,
            dataset.owner,
            dataset.name,
            actions=actions,
            is_internal=config.is_internal,
            expired=_EXPIRED_IN_SECOND,
        )["permission"]

    def _clear_get_permission(self) -> None:
        """Clear the get permission."""
        delattr(self, "_get_permission")

    def _clear_put_permission(self) -> None:
        """Clear the put permission."""
        delattr(self, "_put_permission")

    def _init_put_permission(self) -> Dict[str, Any]:
        """Initialize and return the put permission.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    @property
    def prefix(self) -> str:
        """Return the prefix of the put permission.

        Returns:
            The prefix of the put permission.

        """
        return self._init_put_permission()["prefix"]  # type: ignore[no-any-return]

    def get_object(self, key: str, _allow_retry: bool = True) -> UserResponse:
        """Get the object from graviti.

        Arguments:
            key: The key of the file.
            _allow_retry: Whether requesting the get permission again is allowed.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def put_object(self, key: str, path: Path, _allow_retry: bool = True) -> None:
        """Put the object to OSS.

        Arguments:
            key: The key of the file.
            path: The path of the file.
            _allow_retry: Whether requesting the put permission again is allowed.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class OSSObjectPermissionManager(ObjectPermissionManager):
    """The basic structure of the object permission of the dataset stored in OSS."""

    _RETRY_CODE = {"InvalidAccessKeyId", "AccessDenied"}

    def _init_get_permission(self) -> Dict[str, Any]:
        """Initialize and return the get permission.

        Returns:
            The get permission.

        """
        if self._get_permission is None:
            get_permission: Dict[str, Any] = self._request_permission("GET")
            get_permission["hmac"] = hmac.new(
                get_permission["AccessKeySecret"].encode("utf-8"),
                digestmod=sha1,
            )
            self._get_permission = get_permission
        return self._get_permission

    def _init_put_permission(self) -> Dict[str, Any]:
        """Initialize and return the put permission.

        Returns:
            The put permission.

        """
        if self._put_permission is None:
            put_permission: Dict[str, Any] = self._request_permission("PUT")
            put_permission["hmac"] = hmac.new(
                put_permission["AccessKeySecret"].encode("utf-8"),
                digestmod=sha1,
            )
            self._put_permission = put_permission
        return self._put_permission

    @staticmethod
    def _get_headers(
        permission: Dict[str, Any],
        verb: str,
        key: str,
        mime_type: Optional[str] = None,
    ) -> Dict[str, str]:
        date = convert_datetime_to_gmt(datetime.now(timezone.utc))
        content_type = "" if mime_type is None else mime_type

        signature = (
            f"{verb}\n\n{content_type}\n{date}\n"
            f"x-oss-security-token:{permission['SecurityToken']}\n/{permission['bucket']}/{key}"
        )
        _hmac = permission["hmac"].copy()
        _hmac.update(signature.encode("utf-8"))
        signature = base64.b64encode(_hmac.digest()).decode("utf-8")
        authorization = f"OSS {permission['AccessKeyId']}:{signature}"

        headers = {
            "Authorization": authorization,
            "x-oss-security-token": permission["SecurityToken"],
            "Date": date,
        }
        if mime_type is not None:
            headers["Content-Type"] = mime_type
        return headers

    def get_object(self, key: str, _allow_retry: bool = True) -> UserResponse:
        """Get the object from OSS.

        Arguments:
            key: The key of the file.
            _allow_retry: Whether requesting the get permission again is allowed.

        Raises:
            ResponseError: If post response error.

        Returns:
            The response of OSS get object API.

        """
        permission = self._init_get_permission()
        verb = "GET"

        headers = self._get_headers(permission, verb, key)
        url = f"https://{permission['bucket']}.{permission['endpoint']}/{key}"

        try:
            response = do(verb, url, headers=headers, timeout=config.timeout, stream=True)
            return UserResponse(response)
        except ResponseError as error:
            code = ElementTree.fromstring(error.response.text)[0].text
            if _allow_retry and code in self._RETRY_CODE:
                self._clear_get_permission()
                return self.get_object(key, False)
            raise error from None

    def put_object(self, key: str, path: Path, _allow_retry: bool = True) -> None:
        """Put the object to OSS.

        Arguments:
            key: The key of the file.
            path: The path of the file.
            _allow_retry: Whether requesting the put permission again is allowed.

        Raises:
            ResponseError: If post response error.

        """
        permission = self._init_put_permission()
        verb = "PUT"

        mime_type = mimetypes.guess_type(path)[0]
        headers: Dict[str, Any] = self._get_headers(permission, verb, key, mime_type)
        url = f"https://{permission['bucket']}.{permission['endpoint']}/{key}"

        try:
            with path.open("rb") as fp:
                do(verb, url, headers=headers, data=fp)
        except ResponseError as error:
            code = ElementTree.fromstring(error.response.text)[0].text
            if _allow_retry and code in self._RETRY_CODE:
                self._clear_put_permission()
                self.put_object(key, path, False)
            raise error from None


class AZUREObjectPermissionManager(ObjectPermissionManager):
    """The basic structure of the object permission of the dataset stored in AZURE."""

    _RETRY_CODE = "AuthenticationFailed"

    def _init_get_permission(self) -> Dict[str, Any]:
        """Initialize and return the get permission.

        Returns:
            The get permission.

        """
        if self._get_permission is None:
            self._get_permission = self._request_permission("GET")
        return self._get_permission

    def _init_put_permission(self) -> Dict[str, Any]:
        """Initialize and return the put permission.

        Returns:
            The put permission.

        """
        if self._put_permission is None:
            self._put_permission = self._request_permission("PUT")
        return self._put_permission

    def get_object(self, key: str, _allow_retry: bool = True) -> UserResponse:
        """Get the object from AZURE.

        Arguments:
            key: The key of the file.
            _allow_retry: Whether requesting the get permission again is allowed.

        Raises:
            ResponseError: If post response error.

        Returns:
            The response of AZURE get object API.

        """
        permission = self._init_get_permission()
        verb = "GET"

        url = f"{permission['endpoint_prefix']}/{key}?{permission['sas_param']}"

        try:
            response = do(verb, url, timeout=config.timeout, stream=True)
            return UserResponse(response)
        except ResponseError as error:
            code = ElementTree.fromstring(error.response.text)[0].text
            if _allow_retry and code == self._RETRY_CODE:
                self._clear_get_permission()
                return self.get_object(key, False)
            raise error from None

    def put_object(self, key: str, path: Path, _allow_retry: bool = True) -> None:
        """Put the object to AZURE.

        Arguments:
            key: The key of the file.
            path: The path of the file.
            _allow_retry: Whether requesting the put permission again is allowed.

        Raises:
            ResponseError: If post response error.

        """
        permission = self._init_put_permission()
        verb = "PUT"

        url = f"{permission['endpoint_prefix']}/{key}?{permission['sas_param']}"
        headers = {"x-ms-blob-type": "BlockBlob"}
        mime_type = mimetypes.guess_type(path)[0]
        if mime_type is not None:
            headers["Content-Type"] = mime_type

        try:
            with path.open("rb") as fp:
                do(verb, url, headers=headers, data=fp)
        except ResponseError as error:
            code = ElementTree.fromstring(error.response.text)[0].text
            if _allow_retry and code == self._RETRY_CODE:
                self._clear_put_permission()
                self.put_object(key, path, False)
            raise error from None


class S3ObjectPermissionManager(ObjectPermissionManager):
    """The basic structure of the object permission of the dataset stored in S3."""

    _RETRY_CODE = {"InvalidAccessKeyId", "AccessDenied"}
    _HASHED_PAYLOAD = {"GET": sha256(b"").hexdigest(), "PUT": "UNSIGNED-PAYLOAD"}

    def _init_get_permission(self) -> Dict[str, Any]:
        """Initialize and return the get permission.

        Returns:
            The get permission.

        """
        if self._get_permission is None:
            get_permission: Dict[str, Any] = self._request_permission("GET")
            get_permission["hmac"] = hmac.new(
                f"AWS4{get_permission['AccessKeySecret']}".encode(),
                digestmod=sha256,
            )
            self._get_permission = get_permission
        return self._get_permission

    def _init_put_permission(self) -> Dict[str, Any]:
        """Initialize and return the put permission.

        Returns:
            The put permission.

        """
        if self._put_permission is None:
            put_permission: Dict[str, Any] = self._request_permission("PUT")
            put_permission["hmac"] = hmac.new(
                f"AWS4{put_permission['AccessKeySecret']}".encode(),
                digestmod=sha256,
            )
            self._put_permission = put_permission
        return self._put_permission

    def _get_canonical_request(
        self,
        permission: Dict[str, Any],
        verb: str,
        key: str,
        x_amz_date: str,
    ) -> str:
        hashed_payload = self._HASHED_PAYLOAD[verb]
        canonical_headers = (
            f"host:{permission['bucket']}.{permission['endpoint']}\n"
            f"x-amz-content-sha256:{hashed_payload}\n"
            f"x-amz-date:{x_amz_date}\n"
            f"x-amz-security-token:{permission['SecurityToken']}\n"
        )

        return (
            f"{verb}\n/{key}\n\n{canonical_headers}\nhost;x-amz-content-sha256;x-amz-date;"
            f"x-amz-security-token\n{hashed_payload}"
        )

    @staticmethod
    def _get_string_to_sign(
        permission: Dict[str, Any], simple_date: str, x_amz_date: str, canonical_request: str
    ) -> str:
        scope = f"{simple_date}/{permission['region']}/s3/aws4_request"
        hash_canonical_request = sha256(canonical_request.encode("utf-8")).hexdigest()
        return f"AWS4-HMAC-SHA256\n{x_amz_date}\n{scope}\n{hash_canonical_request}"

    @staticmethod
    def _get_signature(permission: Dict[str, Any], string_to_sign: str, simple_date: str) -> str:
        _hmac = permission["hmac"].copy()
        _hmac.update(simple_date.encode())
        date_region_key = hmac.new(
            _hmac.digest(), f"{permission['region']}".encode(), sha256
        ).digest()
        date_region_service_key = hmac.new(date_region_key, b"s3", sha256).digest()
        signing_key = hmac.new(date_region_service_key, b"aws4_request", sha256).digest()

        return hmac.new(signing_key, string_to_sign.encode("utf-8"), sha256).hexdigest()

    def _get_headers(
        self,
        permission: Dict[str, Any],
        verb: str,
        key: str,
    ) -> Dict[str, str]:
        now = datetime.now(timezone.utc)
        simple_date = f"{now.year:04d}{now.month:02d}{now.day:02}"
        x_amz_date = f"{simple_date}T{now.hour:02d}{now.minute:02d}{now.second:02d}Z"

        canonical_request = self._get_canonical_request(permission, verb, key, x_amz_date)
        string_to_sign = self._get_string_to_sign(
            permission, simple_date, x_amz_date, canonical_request
        )
        signature = self._get_signature(permission, string_to_sign, simple_date)

        credential = (
            f"{permission['AccessKeyId']}/{simple_date}/{permission['region']}/s3/aws4_request"
        )
        authorization = (
            f"AWS4-HMAC-SHA256 Credential={credential}, SignedHeaders=host;"
            f"x-amz-content-sha256;x-amz-date;x-amz-security-token, "
            f"Signature={signature}"
        )

        return {
            "Authorization": authorization,
            "x-amz-security-token": permission["SecurityToken"],
            "x-amz-content-sha256": self._HASHED_PAYLOAD[verb],
            "x-amz-date": x_amz_date,
        }

    def get_object(self, key: str, _allow_retry: bool = True) -> UserResponse:
        """Get the object from S3.

        Arguments:
            key: The key of the file.
            _allow_retry: Whether requesting the get permission again is allowed.

        Raises:
            ResponseError: If post response error.

        Returns:
            The response of S3 get object API.

        """
        permission = self._init_get_permission()
        verb = "GET"

        headers = self._get_headers(permission, verb, key)
        url = f"https://{permission['bucket']}.{permission['endpoint']}/{key}"

        try:
            response = do(verb, url, headers=headers, timeout=config.timeout, stream=True)
            return UserResponse(response)
        except ResponseError as error:
            code = ElementTree.fromstring(error.response.text)[0].text
            if _allow_retry and code in self._RETRY_CODE:
                self._clear_get_permission()
                return self.get_object(key, False)
            raise error from None

    def put_object(self, key: str, path: Path, _allow_retry: bool = True) -> None:
        """Put the object to OSS.

        Arguments:
            key: The key of the file.
            path: The path of the file.
            _allow_retry: Whether requesting the put permission again is allowed.

        Raises:
            ResponseError: If post response error.

        """
        permission = self._init_put_permission()
        verb = "PUT"

        headers: Dict[str, Any] = self._get_headers(permission, verb, key)
        url = f"https://{permission['bucket']}.{permission['endpoint']}/{key}"
        mime_type = mimetypes.guess_type(path)[0]
        if mime_type is not None:
            headers["Content-Type"] = mime_type

        try:
            with path.open("rb") as fp:
                do(verb, url, headers=headers, data=fp)
        except ResponseError as error:
            code = ElementTree.fromstring(error.response.text)[0].text
            if _allow_retry and code in self._RETRY_CODE:
                self._clear_put_permission()
                self.put_object(key, path, False)
            raise error from None



#!/usr/bin/env python3

# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=super-init-not-called
# flake8: noqa

"""Code converting PyArrow schema to Avro Schema."""

from typing import Any, Dict, Type

from graviti.portex.base import PortexType
from graviti.portex.builtin import (
    array,
    binary,
    boolean,
    date,
    enum,
    float32,
    float64,
    int32,
    int64,
    record,
    string,
    time,
    timedelta,
    timestamp,
)


class AvroSchema:
    def __init__(self, name: str, namespace: str, portex_type: PortexType) -> None:
        raise NotImplementedError

    def to_json(self) -> Dict[str, Any]:
        raise NotImplementedError


class AvroField:
    def __init__(
        self,
        type_: AvroSchema,
        name: str,
        *,
        optional: bool = True,
        has_default: bool = False,
        default: Any = None,
    ) -> None:
        self._type = type_
        self._name = name
        self._optional = optional
        self._has_default = has_default
        self._default = default

    def to_json(self) -> Dict[str, Any]:
        if self._optional:
            result: Dict[str, Any] = {
                "name": self._name,
                "type": [
                    "null",
                    self._type.to_json(),
                ],
            }
        else:
            result = {
                "name": self._name,
                "type": self._type.to_json(),
            }

        if self._has_default:
            result["default"] = self._default

        return result


class AvroPrimitiveSchema(AvroSchema):
    _PRIMITIVE_TYPES: Dict[Type[PortexType], str] = {
        int32: "int",
        int64: "long",
        float32: "float",
        float64: "double",
        boolean: "boolean",
        string: "string",
        binary: "bytes",
    }

    def __init__(self, name: str, namespace: str, portex_type: PortexType):
        try:
            self._type = self._PRIMITIVE_TYPES[type(portex_type)]
        except KeyError:
            raise Exception(f"unsupported type {portex_type}") from None

    def to_json(self) -> str:  # type: ignore[override]
        return self._type


class AvroRecord(AvroSchema):
    def __init__(self, name: str, namespace: str, portex_type: record):
        self._name = name
        self._namespace = namespace
        self._fields = [
            AvroField(_avro_schema_creator(key, f"{namespace}.{name}", value), key)
            for key, value in portex_type.items()
        ]

    def to_json(self) -> Dict[str, Any]:
        return {
            "type": "record",
            "name": self._name,
            "namespace": self._namespace,
            "fields": [field.to_json() for field in self._fields],
        }


class AvroArray(AvroSchema):
    def __init__(self, name: str, namespace: str, portex_type: array):
        self._items = _avro_schema_creator("items", f"{namespace}.{name}.items", portex_type.items)

    def to_json(self) -> Dict[str, Any]:
        return {
            "type": "array",
            "items": self._items.to_json(),
            "default": [],
        }


class PortexEnum(AvroSchema):
    def __init__(self, name: str, namespace: str, portex_type: enum):
        self._values = portex_type.values

    def to_json(self) -> Dict[str, Any]:
        min_index, max_index = self._values.index_scope
        avro_type = "int" if min_index >= -2147483648 and max_index <= 2147483647 else "long"

        return {
            "type": avro_type,
            "logicalType": "portex.enum",
            "values": self._values.to_pyobj(),
        }


class PortexDate(AvroSchema):
    def __init__(self, name: str, namespace: str, portex_type: PortexType) -> None:
        pass

    def to_json(self) -> Dict[str, Any]:
        return {
            "type": "int",
            "logicalType": "portex.date",
        }


class PortexTime(AvroSchema):
    _AVRO_TYPES = {
        "s": "int",
        "ms": "int",
        "us": "long",
        "ns": "long",
    }

    def __init__(self, name: str, namespace: str, portex_type: time) -> None:
        self._unit = portex_type.unit

    def to_json(self) -> Dict[str, Any]:
        return {
            "type": self._AVRO_TYPES[self._unit],
            "logicalType": "portex.time",
            "unit": self._unit,
        }


class PortexTimestamp(AvroSchema):
    def __init__(self, name: str, namespace: str, portex_type: timestamp):
        self._unit = portex_type.unit
        self._tz = portex_type.tz

    def to_json(self) -> Dict[str, Any]:
        return {
            "type": "long",
            "logicalType": "portex.timestamp",
            "unit": self._unit,
            "tz": self._tz,
        }


class PortexTimedelta(AvroSchema):
    def __init__(self, name: str, namespace: str, portex_type: timedelta) -> None:
        self._unit = portex_type.unit

    def to_json(self) -> Dict[str, Any]:
        return {
            "type": "long",
            "logicalType": "portex.timedelta",
            "unit": self._unit,
        }


_COMPLEX_TYPE_PROCESSERS: Dict[Type[PortexType], Type[AvroSchema]] = {
    record: AvroRecord,
    array: AvroArray,
    enum: PortexEnum,
    date: PortexDate,
    time: PortexTime,
    timestamp: PortexTimestamp,
    timedelta: PortexTimedelta,
}


def _avro_schema_creator(name: str, namespace: str, portex_type: PortexType) -> AvroSchema:
    builtin_portex_type = portex_type.to_builtin()
    processer = _COMPLEX_TYPE_PROCESSERS.get(type(builtin_portex_type), AvroPrimitiveSchema)
    return processer(name, namespace, builtin_portex_type)


def convert_portex_schema_to_avro(portex_type: record) -> Dict[str, Any]:
    return AvroRecord("root", "cn.graviti.portex", portex_type).to_json()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Paging list related class."""

from functools import partial
from itertools import chain, repeat
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import pyarrow as pa

from graviti.paging.offset import Offsets
from graviti.paging.page import LazyPage, MappedLazyPage, MappedPage, MappedPageBase, Page, PageBase
from graviti.utility import ReprMixin, ReprType

if TYPE_CHECKING:
    from graviti.paging.factory import LazyFactory

_T = TypeVar("_T")
_PLB = TypeVar("_PLB", bound="PagingListBase[Any]")
_PL = TypeVar("_PL", bound="PagingList[Any]")
_MPL = TypeVar("_MPL", bound="MappedPagingList[Any]")
_PPL = TypeVar("_PPL", bound="PyArrowPagingList[Any]")


class PagingListBase(Sequence[_T], ReprMixin):
    """PagingListBase is the base class of the paging list related classes.

    Arguments:
        array: The input sequence.

    """

    _repr_type = ReprType.SEQUENCE

    _array_creator = tuple
    _pages: List[PageBase[_T]]
    _offsets: Offsets

    def __init__(self, iterable: Iterable[_T]) -> None:
        array = self._array_creator(iterable)
        self._init(array)

    def __len__(self) -> int:
        return self._offsets.total_count

    def __iter__(self) -> Iterator[_T]:
        return chain(*self._pages)

    @overload
    def __setitem__(self, index: int, value: _T) -> None:
        ...

    @overload
    def __setitem__(self: _PLB, index: slice, value: Union[Iterable[_T], _PLB]) -> None:
        ...

    def __setitem__(
        self: _PLB, index: Union[int, slice], value: Union[_T, Iterable[_T], _PLB]
    ) -> None:
        if isinstance(index, int):
            self.set_item(index, value)
        elif isinstance(value, self.__class__):
            self.set_slice(index, value)
        else:
            self.set_slice_iterable(index, value)  # type: ignore[arg-type]

    def __delitem__(self, index: Union[int, slice]) -> None:
        if isinstance(index, slice):
            start, stop, step = index.indices(self.__len__())

            if step == 1:
                pass
            elif step == -1:
                start, stop = stop + 1, start + 1
            else:
                ranging: Any = range(start, stop, step)
                if step > 0:
                    ranging = reversed(ranging)

                for i in ranging:
                    self._update_pages(i, i + 1)

                return

        else:
            index = self._make_index_nonnegative(index)
            start = index
            stop = index + 1

        self._update_pages(start, stop)

    @overload
    def __getitem__(self, index: int) -> _T:
        ...

    @overload
    def __getitem__(self: _PLB, index: slice) -> _PLB:
        ...

    def __getitem__(self: _PLB, index: Union[int, slice]) -> Union[_T, _PLB]:
        if isinstance(index, int):
            return self.get_item(index)  # type: ignore[no-any-return]

        return self.get_slice(index)

    def __iadd__(self: _PLB, values: Union[_PLB, Iterable[_T]]) -> _PLB:
        if isinstance(values, self.__class__):
            self.extend(values)
        else:
            self.extend_iterable(values)

        return self

    def _init(self, array: Sequence[_T]) -> None:
        length = len(array)
        self._pages = [Page(array)] if length != 0 else []
        self._offsets = Offsets(length, length)

    def _make_index_nonnegative(self, index: int) -> int:
        return index if index >= 0 else self.__len__() + index

    def _get_slice_positive_step(  # pylint: disable=too-many-locals
        self, start: int, stop: int, step: int
    ) -> List[PageBase[_T]]:
        if start >= stop:
            return []

        _pages = self._pages

        start_i, start_j = self._offsets.get_coordinate(start)
        stop_i, stop_j = self._offsets.get_coordinate(stop - 1)
        stop_j += 1

        if start_i == stop_i:
            return [_pages[start_i].get_slice(start_j, stop_j, step)]

        if step == 1:
            return [
                _pages[start_i].get_slice(start_j),
                *_pages[start_i + 1 : stop_i],
                _pages[stop_i].get_slice(stop=stop_j),
            ]

        start_page = _pages[start_i]
        offset = len(start_page) - start_j
        pages = [start_page.get_slice(start_j, step=step)]

        for page in _pages[start_i + 1 : stop_i]:
            slice_start = -offset % step
            page_length = len(page)

            if slice_start < page_length:
                pages.append(page.get_slice(slice_start, step=step))

            offset += page_length

        stop_page = _pages[stop_i]
        slice_start = -offset % step
        if slice_start < stop_j:
            pages.append(stop_page.get_slice(slice_start, stop_j, step))

        return pages

    def _get_slice_negative_step(  # pylint: disable=too-many-locals
        self, start: int, stop: int, step: int
    ) -> List[PageBase[_T]]:
        if start <= stop:
            return []

        _pages = self._pages

        stop_j: Optional[int]

        start_i, start_j = self._offsets.get_coordinate(start)
        stop_i, stop_j = self._offsets.get_coordinate(stop + 1)
        stop_j = stop_j - 1 if stop_j != 0 else None

        if start_i == stop_i:
            return [_pages[start_i].get_slice(start_j, stop_j, step)]

        if step == -1:
            return [
                _pages[start_i].get_slice(start_j, step=step),
                *(page.get_slice(step=step) for page in _pages[start_i - 1 : stop_i : -1]),
                _pages[stop_i].get_slice(stop=stop_j, step=step),
            ]

        start_page = _pages[start_i]
        offset = start_j + 1
        pages = [start_page.get_slice(start_j, step=step)]

        for page in _pages[start_i - 1 : stop_i : -1]:
            page_length = len(page)
            slice_start = page_length + offset % step - 1

            if slice_start >= 0:
                pages.append(page.get_slice(slice_start, step=step))

            offset += page_length

        stop_page = _pages[stop_i]
        slice_start = len(stop_page) + offset % step - 1

        if slice_start > (stop_j if stop_j is not None else -1):
            pages.append(stop_page.get_slice(slice_start, stop_j, step))

        return pages

    def _update_pages(
        self, start: int, stop: int, pages: Optional[Sequence[PageBase[_T]]] = None
    ) -> None:
        if start >= stop and not pages:
            return

        stop = max(start, stop)

        start_i, start_j = self._offsets.get_coordinate(start)
        stop_i, stop_j = self._offsets.get_coordinate(stop - 1)

        update_pages: List[PageBase[_T]] = []
        update_lengths = []

        left_page = self._pages[start_i].get_slice(stop=start_j)
        left_length = len(left_page)
        if left_length:
            update_pages.append(left_page)
            update_lengths.append(left_length)

        if pages:
            update_pages.extend(pages)
            update_lengths.extend(map(len, pages))

        right_page = self._pages[stop_i].get_slice(stop_j + 1)
        right_length = len(right_page)
        if right_length:
            update_pages.append(right_page)
            update_lengths.append(right_length)

        self._pages[start_i : stop_i + 1] = update_pages
        self._offsets.update(start_i, stop_i, update_lengths)

    def _update_pages_with_step(self: _PLB, start: int, stop: int, step: int, values: _PLB) -> None:
        length = len(values)
        ranging: Any = range(start, stop, step)
        indices: Any = range(length)

        if length != len(ranging):
            raise ValueError(
                f"attempt to assign sequence of size {length} "
                f"to extended slice of size {len(ranging)}"
            )

        if step > 0:
            ranging = reversed(ranging)
            indices = reversed(indices)

        offsets = values._offsets  # pylint: disable=protected-access
        pages = values._pages  # pylint: disable=protected-access

        for i, j in zip(ranging, indices):
            x, y = offsets.get_coordinate(j)
            self._update_pages(i, i + 1, [pages[x][y : y + 1]])

    def get_item(self, index: int) -> _T:
        """Get the element in PagingList at the given index.

        Arguments:
            index: The input index.

        Returns:
            The element at the given index.

        """
        index = self._make_index_nonnegative(index)
        i, j = self._offsets.get_coordinate(index)
        return self._pages[i].get_item(j)

    def get_slice(self: _PLB, index: slice) -> _PLB:
        """Get the sliced PagingList at the given slice.

        Arguments:
            index: The input slice.

        Returns:
            The sliced PagingList at the given slice.

        """
        start, stop, step = index.indices(self.__len__())

        pages = (
            self._get_slice_positive_step(start, stop, step)
            if step > 0
            else self._get_slice_negative_step(start, stop, step)
        )

        offsets = Offsets(0, 0)
        offsets.extend(map(len, pages))

        obj: _PLB = object.__new__(self.__class__)
        # pylint: disable=protected-access
        obj._pages = pages
        obj._offsets = offsets

        return obj

    def set_item(self, index: int, value: _T) -> None:
        """Update the element value in PagingList at the given index.

        Arguments:
            index: The element index.
            value: The value needs to be set into the PagingList.

        """
        index = self._make_index_nonnegative(index)
        page = Page(self._array_creator((value,)))
        self._update_pages(index, index + 1, [page])

    def set_slice(self: _PLB, index: slice, values: _PLB) -> None:
        """Update the element values at the given slice with input PagingList.

        Arguments:
            index: The element slice.
            values: The PagingList which contains the elements to be set.

        Raises:
            ValueError: When the input size mismatches with the slice size (when step != 1).

        """
        start, stop, step = index.indices(self.__len__())

        if step == 1:
            self._update_pages(start, stop, values._pages)  # pylint: disable=protected-access
            return

        if step == -1:
            start, stop = stop + 1, max(start, stop) + 1
            if len(values) != stop - start:
                raise ValueError(
                    f"attempt to assign sequence of size {len(values)} "
                    f"to extended slice of size {stop - start}"
                )

            self._update_pages(
                start,
                stop,
                [
                    page.get_slice(step=-1)
                    for page in reversed(values._pages)  # pylint: disable=protected-access
                ],
            )
            return

        self._update_pages_with_step(start, stop, step, values)

    def set_slice_iterable(self, index: slice, values: Iterable[_T]) -> None:
        """Update the element values in PagingList at the given slice with iterable object.

        Arguments:
            index: The element slice.
            values: The iterable object which contains the elements to be set.

        Raises:
            ValueError: When the assign input size mismatches with the slice size (when step != 1).

        """
        start, stop, step = index.indices(self.__len__())

        if step == 1:
            array = self._array_creator(values)
            self._update_pages(start, stop, [Page(array)] if len(array) != 0 else None)
            return

        if step == -1:
            try:
                values = reversed(values)  # type: ignore[call-overload]
            except TypeError:
                values = reversed(list(values))

            array = self._array_creator(values)

            start, stop = stop + 1, max(start, stop) + 1
            if len(array) != stop - start:
                raise ValueError(
                    f"attempt to assign sequence of size {len(array)} "
                    f"to extended slice of size {stop - start}"
                )

            self._update_pages(start, stop, [Page(array)] if len(array) != 0 else None)
            return

        self._update_pages_with_step(start, stop, step, self.__class__(self._array_creator(values)))

    def extend(self: _PLB, values: _PLB) -> None:
        """Extend PagingList by appending elements from another PagingList.

        Arguments:
            values: The PagingList which contains the elements to be extended.

        """
        pages = values._pages  # pylint: disable=protected-access
        self._offsets.extend(map(len, pages))
        self._pages.extend(pages)

    def extend_iterable(self, values: Iterable[_T]) -> None:
        """Extend PagingList by appending elements from the iterable.

        Arguments:
            values: Elements to be extended into the PagingList.

        """
        page = Page(self._array_creator(values))
        self._offsets.extend((len(page),))
        self._pages.append(page)

    def extend_nulls(self, size: int) -> None:
        """Extend PagingList by appending nulls.

        Arguments:
            size: The size of the nulls to be extended.

        """
        page = Page(self._array_creator(repeat(None, size)))
        self._offsets.extend((len(page),))
        self._pages.append(page)  # type: ignore[arg-type]

    def copy(self: _PLB) -> _PLB:
        """Return a copy of the paging list.

        Returns:
            A copy of the paging list.

        """
        obj: _PLB = object.__new__(self.__class__)
        # pylint: disable=protected-access
        obj._pages = self._pages.copy()
        obj._offsets = self._offsets.copy()
        return obj


class PagingList(PagingListBase[_T]):
    """PagingList is a list composed of multiple lists (pages)."""

    @classmethod
    def from_factory(
        cls: Type[_PL],
        factory: "LazyFactory",
        keys: Tuple[str, ...],
        mapper: Callable[[Any], _T],
    ) -> _PL:
        """Create PagingList from LazyFactory.

        Arguments:
            factory: The parent :class:`LazyFactory` instance.
            keys: The keys to access the array from factory.
            mapper: A callable object to convert every item in the pyarrow array.

        Returns:
            The PagingList instance created from given factory.

        """
        obj: _PL = object.__new__(cls)

        def get_array(pos: int, keys: Tuple[str, ...]) -> Tuple[Any, ...]:
            return tuple(map(mapper, factory.get_array(pos, keys)))

        obj._pages = [
            LazyPage(length, partial(get_array, pos, keys))
            for pos, length in enumerate(factory.get_page_lengths())
        ]
        obj._offsets = factory.get_offsets()

        return obj


class MappedPagingList(PagingListBase[_T]):
    """MappedPagingList is a list composed of multiple mapped pages."""

    _pages: List[MappedPageBase[_T]]  # type: ignore[assignment]

    def _init(self, array: Sequence[_T]) -> None:
        length = len(array)
        self._pages = [MappedPage(array)] if length != 0 else []
        self._offsets = Offsets(length, length)

    @classmethod
    def from_array(
        cls: Type[_MPL],
        array: Sequence[_T],
        mapper: Callable[[Any], _T],
    ) -> _MPL:
        """Create MappedPagingList from the source array.

        Arguments:
            array: The source array of the paging list.
            mapper: A callable object to convert every item in the pyarrow array.

        Returns:
            The PagingList instance created from the given array.

        """
        length = len(array)
        obj: _MPL = object.__new__(cls)

        obj._pages = [MappedLazyPage(length, lambda: array, mapper)]
        obj._offsets = Offsets(length, length)

        return obj

    @classmethod
    def from_factory(
        cls: Type[_MPL],
        factory: "LazyFactory",
        keys: Tuple[str, ...],
        mapper: Callable[[Any], _T],
    ) -> _MPL:
        """Create MappedPagingList from LazyFactory.

        Arguments:
            factory: The parent :class:`LazyFactory` instance.
            keys: The keys to access the array from factory.
            mapper: A callable object to convert every item in the pyarrow array.

        Returns:
            The PagingList instance created from given factory.

        """
        obj: _MPL = object.__new__(cls)

        obj._pages = [
            MappedLazyPage(length, partial(factory.get_array, pos, keys), mapper)
            for pos, length in enumerate(factory.get_page_lengths())
        ]
        obj._offsets = factory.get_offsets()

        return obj

    def copy(  # type: ignore[override] # pylint: disable=arguments-differ
        self: _MPL,
        copier: Callable[[_T], _T],
        mapper: Callable[[Any], _T],
    ) -> _MPL:
        """Return a copy of the paging list.

        Arguments:
            copier: A callable object to convert loaded items in the source page to the copied page.
            mapper: The mapper of the new mapped page.

        Returns:
            A copy of the paging list.

        """
        obj: _MPL = object.__new__(self.__class__)
        # pylint: disable=protected-access
        obj._offsets = self._offsets.copy()
        obj._pages = [page.copy(copier, mapper) for page in self._pages]
        return obj


class PyArrowPagingList(PagingListBase[_T]):
    """PyArrowPagingList is a list composed of multiple pyarrow arrays (pages).

    Arguments:
        array: The input pyarrow array.

    """

    _array_creator = pa.array
    _patype: pa.DataType

    def _init(self, array: pa.Array) -> None:
        super()._init(array)
        self._patype = array.type
        self._array_creator = partial(pa.array, type=array.type)

    @classmethod
    def from_pyarrow(cls: Type[_PPL], array: pa.Array) -> _PPL:
        """Create PyArrowPagingList from pyarrow array.

        Arguments:
            array: The input pyarrow array.

        Returns:
            The PyArrowPagingList instance created from given pyarrow array.

        """
        obj: _PPL = object.__new__(cls)
        obj._init(array)

        return obj

    @classmethod
    def from_factory(
        cls: Type[_PPL], factory: "LazyFactory", keys: Tuple[str, ...], patype: pa.DataType
    ) -> _PPL:
        """Create PyArrowPagingList from LazyFactory.

        Arguments:
            factory: The parent :class:`LazyFactory` instance.
            keys: The keys to access the array from factory.
            patype: The pyarrow DataType of the elements in the list.

        Returns:
            The PyArrowPagingList instance created from given factory.

        """
        obj: _PPL = object.__new__(cls)
        obj._pages = [
            LazyPage(length, partial(factory.get_array, pos, keys))
            for pos, length in enumerate(factory.get_page_lengths())
        ]
        obj._offsets = factory.get_offsets()
        obj._patype = patype

        return obj

    def get_slice(self: _PPL, index: slice) -> _PPL:
        """Get the sliced PyArrowPagingList at the given slice.

        Arguments:
            index: The input slice.

        Returns:
            The sliced PyArrowPagingList at the given slice.

        """
        obj = super().get_slice(index)
        obj._patype = self._patype  # pylint: disable=protected-access

        return obj

    def set_slice(self: _PPL, index: slice, values: _PPL) -> None:
        """Update the element values at the given slice with input PyArrowPagingList.

        Arguments:
            index: The element slice.
            values: The PyArrowPagingList which contains the elements to be set.

        Raises:
            ArrowTypeError: When two pyarrow types mismatch.

        """
        # pylint: disable=protected-access
        if values._patype != self._patype:
            raise pa.ArrowTypeError(
                f"Can not set a '{self._patype}' list with a '{values._patype}' list"
            )

        super().set_slice(index, values)

    def extend(self: _PPL, values: _PPL) -> None:
        """Extend PyArrowPagingList by appending elements from another PyArrowPagingList.

        Arguments:
            values: The PyArrowPagingList which contains the elements to be extended.

        Raises:
            ArrowTypeError: When two pyarrow types mismatch.

        """
        # pylint: disable=protected-access
        if values._patype != self._patype:
            raise pa.ArrowTypeError(
                f"Can not extend a '{self._patype}' list with a '{values._patype}' list"
            )

        super().extend(values)

    def extend_nulls(self, size: int) -> None:
        """Extend PyArrowPagingList by appending nulls.

        Arguments:
            size: The size of the nulls to be extended.

        """
        page = Page(pa.nulls(size, self._patype))
        self._offsets.extend((len(page),))
        self._pages.append(page)

    def copy(self: _PPL) -> _PPL:
        """Return a copy of the paging list.

        Returns:
            A copy of the paging list.

        """
        obj = super().copy()
        # pylint: disable=protected-access
        obj._array_creator = self._array_creator
        obj._patype = self._patype
        return obj

    def to_pyarrow(self) -> pa.ChunkedArray:
        """Convert the paging list to pyarrow ChunkedArray.

        Returns:
            The pyarrow ChunkedArray.

        """
        return pa.chunked_array((page.get_array() for page in self._pages), self._patype)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""search module."""

from graviti.dataframe.sql.array import DataFrame
from graviti.dataframe.sql.scalar import RowSeries

__all__ = ["DataFrame", "RowSeries"]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Template factory releated classes."""


from copy import deepcopy
from itertools import groupby
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from graviti.portex import ptype as PTYPE
from graviti.portex.base import PortexRecordBase, PortexType
from graviti.portex.field import ConnectedFields, Fields, FrozenFields
from graviti.portex.package import Imports

_C = TypeVar("_C", str, float, bool, None)
_CFF = TypeVar("_CFF", bound="ConnectedFieldsFactory")


class Factory:
    """The base class of the template factory."""

    is_unpack: bool = False
    keys: Dict[str, Any]

    def __call__(self, _: Dict[str, Any]) -> Any:
        """Apply the input arguments to the template.

        Arguments:
            _: The input arguments.

        """
        ...


class FrozenFieldsFactory(Factory):
    """The factory for FrozenFields.

    Arguments:
        decl: The decalaration of frozen fields.
        imports: The :class:`Imports` instance to specify the import scope of the fields.

    """

    def __init__(self, decl: Iterable[Dict[str, Any]], imports: Imports) -> None:
        self._factories = [FieldFactory(field, imports) for field in decl]

    def __call__(self, kwargs: Dict[str, Any]) -> FrozenFields:
        """Apply the input arguments to the FrozenFields factory.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The applied FrozenFields.

        """
        return FrozenFields(
            filter(
                bool,
                (factory(kwargs) for factory in self._factories),  # type: ignore[misc]
            )
        )


class FrozenFieldsFactoryWrapper(Factory):
    """The factory for FrozenFields which needs kwargs transformed.

    Arguments:
        factory: The factory of frozen fields.
        kwargs_transformer: The method to transform the kwargs to the kwargs of base type.

    """

    def __init__(
        self,
        factory: Union[FrozenFieldsFactory, "FrozenFieldsFactoryWrapper"],
        kwargs_transformer: Callable[..., Dict[str, Any]],
    ) -> None:
        self._factory = factory
        self._kwargs_transformer = kwargs_transformer

    def __call__(self, kwargs: Dict[str, Any]) -> FrozenFields:
        """Apply the input arguments to the base FrozenFields factory.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The applied FrozenFields.

        """
        return self._factory(self._kwargs_transformer(kwargs))


UnionFieldsFactory = Union[
    "VariableFactory", "ConstantFactory", FrozenFieldsFactory, FrozenFieldsFactoryWrapper
]


class ConnectedFieldsFactory:
    """The factory for ConnectedFields.

    Arguments:
        decl: A dict which indicates a portex type.
        class_: The base type.
        imports: The :class:`Imports` instance to specify the import scope of the template.
        kwargs_transformer: The method to transform the kwargs to the kwargs of base type.

    """

    _factories: List[UnionFieldsFactory]

    def __init__(
        self,
        decl: Dict[str, Any],
        class_: Type[PortexRecordBase],
        imports: Imports,
        kwargs_transformer: Callable[..., Dict[str, Any]],
    ) -> None:
        factories: List[UnionFieldsFactory] = []
        for base_factory in class_._fields_factory._factories:
            if not isinstance(base_factory, VariableFactory):
                if isinstance(base_factory, ConstantFactory):
                    factories.append(base_factory)
                else:
                    factories.append(FrozenFieldsFactoryWrapper(base_factory, kwargs_transformer))
                continue

            key = base_factory.key
            if key not in decl:
                factories.append(
                    ConstantFactory(FrozenFields(deepcopy(class_.params[key].default)))
                )
                continue

            fields_decl = decl[key]
            if isinstance(fields_decl, str) and fields_decl.startswith("$"):
                factories.append(VariableFactory(fields_decl[1:]))
                continue

            groups = groupby(fields_decl, lambda x: isinstance(x, str) and x.startswith("+$"))
            for is_mutable, fields in groups:
                if is_mutable:
                    for unpack_fields in fields:
                        factories.append(VariableFactory(unpack_fields[2:]))
                else:
                    factories.append(FrozenFieldsFactory(fields, imports))

        self._factories = factories

    def __call__(self, kwargs: Dict[str, Any]) -> ConnectedFields:
        """Apply the input arguments to the ConnectedFields factory.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The applied ConnectedFields.

        """
        return ConnectedFields(factory(kwargs) for factory in self._factories)

    @classmethod
    def from_parameter_name(cls: Type[_CFF], name: str) -> _CFF:
        """Create ConnectedFieldsFactory for Fields with the given parameter name.

        Arguments:
            name: The parameter name of the input fields.

        Returns:
            The created ConnectedFieldsFactory.

        """
        obj: _CFF = object.__new__(cls)
        obj._factories = [VariableFactory(name)]
        return obj


class TypeFactory(Factory):
    """The template factory for portex type.

    Arguments:
        decl: A dict which indicates a portex type.

    """

    # After removing utility/pyarrow.py in pull request #420,
    # mypy raises error: Cannot determine type of "class_"  [has-type].
    # This is a workaround for the above mentioned error.
    class_: Type[PortexType]

    def __init__(self, decl: Dict[str, Any], imports: Imports) -> None:
        class_ = imports[decl["type"]]

        factories = {}
        keys = {}

        for name, parameter in class_.params.items():
            try:
                value = decl[name]
            except KeyError as error:
                if parameter.required:
                    raise KeyError(f"Parameter '{name}' is required") from error
                continue

            factory = factory_creator(value, imports, parameter.ptype)
            if factory.is_unpack:
                raise ValueError("Use array unpack in object value is not allowed")
            factories[name] = factory
            keys.update(factory.keys)

        if "+" in decl:
            unpack_factory = mapping_unpack_factory_creator(decl["+"], PTYPE.Mapping)
            keys.update(unpack_factory.keys)
            self._unpack_factory = unpack_factory

        self._factories = factories
        self.keys = keys
        self.class_ = class_

    def __call__(self, kwargs: Dict[str, Any]) -> PortexType:
        """Apply the input arguments to the type template.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The applied Portex type.

        """
        return self.class_(**self.transform_kwargs(kwargs))  # type: ignore[call-arg]

    def transform_kwargs(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Transform the keyword arguments to what the base type needs.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The transformed keyword arguments.

        """
        type_kwargs: Dict[str, Any] = (
            self._unpack_factory(kwargs) if hasattr(self, "_unpack_factory") else {}
        )
        type_kwargs.update({key: factory(kwargs) for key, factory in self._factories.items()})
        return type_kwargs


class ConstantFactory(Factory, Generic[_C]):
    """The template factory for a constant.

    Arguments:
        decl: The constant to be created by the factory.

    """

    def __init__(self, decl: _C) -> None:
        self._constant: _C = decl
        self.keys: Dict[str, Any] = {}

    def __call__(self, _: Dict[str, Any]) -> _C:
        """Get the constant stored in the factory.

        Arguments:
            _: The input arguments.

        Returns:
            The constant stored in the factory.

        """
        return self._constant


class VariableFactory(Factory):
    """The template factory for a variable.

    Arguments:
        decl: The parameter name of the variable.
        ptype: The parameter type.

    """

    def __init__(self, decl: str, ptype: PTYPE.PType = PTYPE.Any, is_unpack: bool = False) -> None:
        self.key = decl
        self.keys = {decl: ptype}
        self.is_unpack = is_unpack

    def __call__(self, kwargs: Dict[str, Any]) -> Any:
        """Apply the input arguments to the variable template.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The applied variable.

        """
        return kwargs[self.key]


class ListFactory(Factory):
    """The template factory for a list.

    Arguments:
        decl: A list template.
        ptype: The parameter type of the list.

    """

    def __init__(self, decl: List[Any], ptype: PTYPE.PType = PTYPE.Any) -> None:
        factories = []
        keys = {}
        for value in decl:
            factory = factory_creator(value, None, ptype)
            factories.append(factory)
            keys.update(factory.keys)

        self._factories = factories
        self.keys = keys

    def __call__(self, kwargs: Dict[str, Any]) -> List[Any]:
        """Apply the input arguments to the list template.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The applied list.

        """
        return list(_generate_factory_items(self._factories, kwargs))


class DictFactory(Factory):
    """The template factory for a dict.

    Arguments:
        decl: A dict template.
        ptype: The parameter type of the dict.

    """

    def __init__(self, decl: Dict[str, Any], ptype: PTYPE.PType = PTYPE.Any) -> None:
        factories = {}
        keys = {}

        for key, value in decl.items():
            if key == "+":
                unpack_factory = mapping_unpack_factory_creator(value, PTYPE.Mapping)
                keys.update(unpack_factory.keys)
                self._unpack_factory = unpack_factory
                continue

            factory = factory_creator(value, None, ptype)
            if factory.is_unpack:
                raise ValueError("Use array unpack in object value is not allowed")
            factories[key] = factory
            keys.update(factory.keys)

        self._factories = factories
        self.keys = keys

    def __call__(self, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Apply the input arguments to the dict template.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The applied dict.

        """
        items: Dict[str, Any] = (
            self._unpack_factory(kwargs) if hasattr(self, "_unpack_factory") else {}
        )
        items.update({key: factory(kwargs) for key, factory in self._factories.items()})
        return items


class FieldFactory(Factory):
    """The template factory for a tuple of name and PortexType.

    Arguments:
        decl: A dict which indicates a tuple of name and PortexType.

    """

    def __init__(self, decl: Dict[str, Any], imports: Imports) -> None:
        self.creator: Callable[..., Tuple[str, PortexType]]
        if "+" in decl and len(decl) == 1:
            raise ValueError(
                "Use object unpack for entire record field is not allowed. "
                "Please use list unpack for fields"
            )

        item = decl.copy()
        keys = {}

        condition = string_factory_creator(item.pop("exist_if", True))
        keys.update(condition.keys)

        name_factory = string_factory_creator(item.pop("name"), PTYPE.String)
        type_factory = type_factory_creator(item, imports)

        keys.update(name_factory.keys)
        keys.update(type_factory.keys)

        self._condition = condition
        self._name_factory = name_factory
        self._type_factory = type_factory
        self.keys = keys

    def __call__(self, kwargs: Dict[str, Any]) -> Optional[Tuple[str, PortexType]]:
        """Apply the input arguments to the template.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The applied tuple of name and PortexType.

        """
        if not self._condition(kwargs):
            return None

        return self._name_factory(kwargs), self._type_factory(kwargs)


class FieldsFactory(Factory):
    """The template factory for a ``Fields``.

    Arguments:
        decl: A list which indicates a ``Fields``.

    """

    def __init__(self, decl: List[Union[Dict[str, Any], str]], imports: Imports) -> None:
        self._factories = []
        keys = {}

        for item in decl:
            if isinstance(item, dict):
                factory: Factory = FieldFactory(item, imports)
            elif isinstance(item, str) and item.startswith("+$"):
                factory = VariableFactory(item[2:], PTYPE.Fields, True)
            else:
                raise ValueError("The items of fields can only be object or list unpack parameter")

            keys.update(factory.keys)
            self._factories.append(factory)

        self.keys = keys

    def __call__(self, kwargs: Dict[str, Any]) -> Fields:
        """Apply the input arguments to the ``Fields`` template.

        Arguments:
            kwargs: The input arguments.

        Returns:
            The applied ``Fields``.

        """
        return Fields(
            filter(
                bool,
                _generate_factory_items(self._factories, kwargs, lambda x: x.items()),
            )
        )


def _generate_factory_items(
    factories: Iterable[Factory],
    kwargs: Dict[str, Any],
    unpack_item_processer: Callable[[Any], Any] = lambda x: x,
) -> Iterator[Any]:
    for factory in factories:
        item = factory(kwargs)
        if factory.is_unpack:
            yield from unpack_item_processer(item)
        else:
            yield item


def mapping_unpack_factory_creator(decl: str, ptype: PTYPE.PType) -> VariableFactory:
    """Check the object unpack grammar and returns the corresponding factory.

    Arguments:
        decl: The parameter decalaration.
        ptype: The parameter type of the input.

    Raises:
        ValueError: When the object unpack grammar is incorrect.

    Returns:
        A ``VariableFactory`` instance according to the input.

    """
    if not decl.startswith("$"):
        raise ValueError("The decl does not have the correct object unpack grammar")

    return VariableFactory(decl[1:], ptype)


def type_factory_creator(
    decl: Dict[str, Any], imports: Imports
) -> Union[TypeFactory, VariableFactory]:
    """Check the input and returns the corresponding type factory.

    Arguments:
        decl: A dict which indicates a portex type or has object unpack grammar.
        imports: The :class:`Imports` instance to specify the import scope of the template.

    Raises:
        ValueError: When setting the type name as a parameter.

    Returns:
        A ``TypeFactory`` or a ``VariableFactory`` instance.

    """
    if "type" not in decl:
        return mapping_unpack_factory_creator(decl["+"], PTYPE.PortexType)

    if decl["type"].startswith("$"):
        raise ValueError(
            "Setting the type name as a parameter is not allowed. Please use object unpack grammar"
        )

    return TypeFactory(decl, imports)


def string_factory_creator(
    decl: str, ptype: PTYPE.PType = PTYPE.Any
) -> Union[VariableFactory, ConstantFactory[str]]:
    """Check whether the input string is variable and returns the corresponding factory.

    Arguments:
        decl: A string which indicates a constant or a variable.
        ptype: The parameter type of the string.

    Returns:
        A ``VariableFactory`` or a ``ConstantFactory`` instance according to the input.

    """
    if isinstance(decl, str) and decl.startswith("$"):
        return VariableFactory(decl[1:], ptype)

    return ConstantFactory(decl)


def factory_creator(  # pylint: disable=too-many-return-statements
    decl: Any, imports: Optional[Imports], ptype: PTYPE.PType = PTYPE.Any
) -> Factory:
    """Check input type and returns the corresponding factory.

    Arguments:
        decl: A template which indicates any Portex object.
        imports: The :class:`Imports` instance to specify the import scope of the template.
        ptype: The parameter type of the input.

    Returns:
        A ``Factory`` instance according the input.

    """
    if isinstance(decl, str):
        if decl.startswith("$"):
            return VariableFactory(decl[1:], ptype)

        if decl.startswith("+$"):
            return VariableFactory(decl[2:], PTYPE.Array, True)

    if ptype == PTYPE.PortexType:
        assert isinstance(decl, dict)
        assert imports is not None
        return type_factory_creator(decl, imports)

    if ptype == PTYPE.Fields:
        assert isinstance(decl, list)
        assert imports is not None
        return FieldsFactory(decl, imports)

    if isinstance(decl, list):
        return ListFactory(decl, ptype)

    if isinstance(decl, dict):
        return DictFactory(decl, ptype)

    return ConstantFactory(decl)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Dataset and DatasetManager."""

import logging
from enum import Enum
from typing import Any, Dict, Generator, Optional, Tuple, TypeVar

from graviti.dataframe import DataFrame
from graviti.exception import ResourceNameError, StatusError
from graviti.manager.branch import Branch, BranchManager
from graviti.manager.commit import Commit, CommitManager
from graviti.manager.common import LIMIT
from graviti.manager.draft import DraftManager
from graviti.manager.lazy import LazyPagingList
from graviti.manager.permission import (
    AZUREObjectPermissionManager,
    OSSObjectPermissionManager,
    S3ObjectPermissionManager,
)
from graviti.manager.tag import Tag, TagManager
from graviti.openapi import (
    create_dataset,
    delete_dataset,
    get_dataset,
    get_revision,
    list_datasets,
    update_dataset,
)
from graviti.utility import (
    ReprMixin,
    ReprType,
    UserMutableMapping,
    check_type,
    convert_iso_to_datetime,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
logger.addHandler(handler)


class RevisionType(Enum):
    """RevisionType is an enumeration type including "BRANCH", "COMMIT" and "TAG"."""

    BRANCH = Branch
    COMMIT = Commit
    TAG = Tag


class ObjectPermissionManagerType(Enum):
    """ObjectPermissionManagerType is an enumeration type including "OSS", "S3" and "AZURE"."""

    OSS = OSSObjectPermissionManager
    S3 = S3ObjectPermissionManager
    AZURE = AZUREObjectPermissionManager


class Dataset(  # pylint: disable=too-many-instance-attributes
    UserMutableMapping[str, DataFrame], ReprMixin
):
    """This class defines the basic concept of the dataset on Graviti.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        response: The response of the OpenAPI associated with the dataset::

                {
                    "id": <str>
                    "name": <str>
                    "alias": <str>
                    "default_branch": <str>
                    "commit_id": <Optional[str]>
                    "cover_url": <str>
                    "created_at": <str>
                    "updated_at": <str>
                    "owner": <str>
                    "is_public": <bool>
                    "config": <str>
                    "backend_type": <str>
                }

    Attributes:
        dataset_id: Dataset ID.
        name: The name of the dataset, unique for a user.
        alias: Dataset alias.
        default_branch: The default branch of dataset.
        commit_id: The commit ID of the dataset.
        created_at: The time when the dataset was created.
        updated_at: The time when the dataset was last modified.
        owner: The owner of the dataset.
        is_public: Whether the dataset is public.
        config: The config of dataset.
        backend_type: The backend type of dataset storage.

    """

    _T = TypeVar("_T", bound="Dataset")

    _repr_type = ReprType.INSTANCE

    _repr_attrs: Tuple[str, ...] = (
        "alias",
        "default_branch",
        "created_at",
        "updated_at",
        "is_public",
        "config",
        "branch",
    )

    def __init__(self, access_key: str, url: str, response: Dict[str, Any]) -> None:
        self.access_key = access_key
        self.url = url
        self._dataset_id = response["id"]
        self.name = response["name"]
        self.alias = response["alias"]
        self.default_branch = response["default_branch"]
        self.created_at = convert_iso_to_datetime(response["created_at"])
        self.updated_at = convert_iso_to_datetime(response["updated_at"])
        self.owner = response["owner"]
        self.is_public = response["is_public"]
        self.config = response["config"]

        backend_type = response["backend_type"]
        self.object_permission_manager = ObjectPermissionManagerType[backend_type].value(self)

        self._data: Commit = Branch(self, response["default_branch"], response["commit_id"])

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.owner}/{self.name}")'

    @property
    def HEAD(self) -> Commit:  # pylint: disable=invalid-name
        """Return the current branch or commit.

        Returns:
            The current branch or commit.

        """
        return self._data

    @property
    def branches(self) -> BranchManager:
        """Get class :class:`~graviti.manager.branch.BranchManager` instance.

        Returns:
            Required :class:`~graviti.manager.branch.BranchManager` instance.

        """
        return BranchManager(self)

    @property
    def drafts(self) -> DraftManager:
        """Get class :class:`~graviti.manager.draft.DraftManager` instance.

        Returns:
            Required :class:`~graviti.manager.draft.DraftManager` instance.

        """
        return DraftManager(self)

    @property
    def commits(self) -> CommitManager:
        """Get class :class:`~graviti.manager.commit.CommitManager` instance.

        Returns:
            Required :class:`~graviti.manager.commit.CommitManager` instance.

        """
        return CommitManager(self)

    @property
    def tags(self) -> TagManager:
        """Get class :class:`~graviti.manager.tag.TagManager` instance.

        Returns:
            Required :class:`~graviti.manager.tag.TagManager` instance.

        """
        return TagManager(self)

    def checkout(self, revision: str) -> None:
        """Checkout to a commit.

        Arguments:
            revision: The information to locate the specific commit, which can be the commit id,
                the branch, or the tag.

        """
        response = get_revision(
            self.access_key,
            self.url,
            self.owner,
            self.name,
            revision=revision,
        )

        response["name"] = revision
        revision_type = RevisionType[response["type"]]

        self._data = revision_type.value.from_response(self, response)

    def edit(
        self,
        *,
        name: Optional[str] = None,
        alias: Optional[str] = None,
        default_branch: Optional[str] = None,
    ) -> None:
        """Update the meta data of the dataset.

        Arguments:
            name: The new name of the dataset.
            alias: The new alias of the dataset.
            default_branch: The new default branch of the dataset.

        """
        response = update_dataset(
            self.access_key,
            self.url,
            self.owner,
            self.name,
            name=name,
            alias=alias,
            default_branch=default_branch,
        )
        self.name = response["name"]
        self.alias = response["alias"]
        self.default_branch = response["default_branch"]
        self.updated_at = convert_iso_to_datetime(response["updated_at"])

    def commit(
        self, title: str, description: Optional[str] = None, jobs: int = 8, quiet: bool = False
    ) -> None:
        """Create, upload and commit the draft to push the local dataset to Graviti.

        Arguments:
            title: The commit title.
            description: The commit description.
            jobs: The number of the max workers in multi-thread upload, the default is 8.
            quiet: Set to True to stop showing the upload process bar.

        Raises:
            StatusError: When the HEAD of the dataset is not a branch.
            StatusError: When the dataset has no modifications.

        """
        head = self.HEAD
        if not isinstance(head, Branch):
            raise StatusError(
                "It is not allowed to commit a dataset whose HEAD is not a branch. "
                "Please checkout a branch first"
            )
        modified = head.operations or any(df.operations for df in head.values())
        if not modified:
            raise StatusError("It is not allowed to commit a dataset without any modifications")

        # pylint: disable=protected-access
        draft = self.drafts.create(title, description, head.name)
        logger.info("%s created successfully", draft._repr_head())

        head._upload_to_draft(draft.number, jobs, quiet)
        logger.info("%s uploaded successfully", draft._repr_head())

        branch = draft.commit(title, description)
        logger.info("%s committed successfully", draft._repr_head())
        logger.info("The HEAD of the dataset after commit: \n%s", branch)


class DatasetManager:
    """This class defines the operations on the dataset on Graviti.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.

    """

    def __init__(self, access_key: str, url: str, owner: str) -> None:
        self.access_key = access_key
        self.url = url
        self.owner = owner

    def _generate(self, offset: int, limit: int) -> Generator[Dataset, None, int]:
        access_key = self.access_key
        url = self.url
        response = list_datasets(access_key, url, limit=limit, offset=offset)

        for item in response["datasets"]:
            yield Dataset(access_key, url, item)

        return response["total_count"]  # type: ignore[no-any-return]

    def create(
        self,
        name: str,
        alias: str = "",
        config: Optional[str] = None,
    ) -> Dataset:
        """Create a Graviti dataset with given name.

        Arguments:
            name: The name of the dataset, unique for a user.
            alias: Alias of the dataset, default is "".
            config: The auth storage config name.

        Returns:
            The created :class:`~graviti.manager.dataset.Dataset` instance.

        """
        response = create_dataset(
            self.access_key,
            self.url,
            name=name,
            alias=alias,
            config=config,
            with_draft=False,
        )

        return Dataset(self.access_key, self.url, response)

    def get(self, dataset: str) -> Dataset:
        """Get a Graviti dataset with given name.

        Arguments:
            dataset: The name of the dataset, unique for a user.

        Returns:
            The requested :class:`~graviti.manager.dataset.Dataset` instance.

        Raises:
            ResourceNameError: When the required dataset does not exist.

        """
        check_type("dataset", dataset, str)
        if not dataset:
            raise ResourceNameError("dataset", dataset)

        response = get_dataset(self.access_key, self.url, owner=self.owner, dataset=dataset)

        return Dataset(self.access_key, self.url, response)

    def list(self) -> LazyPagingList[Dataset]:
        """List Graviti datasets.

        Returns:
            The LazyPagingList of :class:`~graviti.manager.dataset.Dataset` instances.

        """
        return LazyPagingList(self._generate, LIMIT)

    def delete(self, name: str) -> None:
        """Delete a Graviti dataset with given name.

        Arguments:
            name: The name of the dataset, unique for a user.

        """
        check_type("name", name, str)
        delete_dataset(self.access_key, self.url, self.owner, name)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of request related tools."""

import logging
import os
from collections import defaultdict
from concurrent.futures import FIRST_EXCEPTION, ThreadPoolExecutor, wait
from typing import Any, Callable, DefaultDict, Iterable, Optional, TypeVar

import urllib3
from requests import Session
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException, StreamConsumedError
from requests.models import PreparedRequest, Response
from urllib3.util.retry import Retry

from graviti.exception import ResponseError
from graviti.utility.log import RequestLogging, ResponseLogging

logger = logging.getLogger(__name__)


_CHUNK_SIZE = 8 * 1024


def _get_allowed_methods_keyword() -> str:
    splits = urllib3.__version__.split(".", 2)
    major = int(splits[0])
    minor = int(splits[1])
    return "allowed_methods" if (major, minor) >= (1, 26) else "method_whitelist"


# check the version of urllib3 and choose the correct keyword for "allowed_methods" in "Retry"
_ALLOWED_METHODS = _get_allowed_methods_keyword()


class Config:
    """This is a base class defining the concept of Request Config.

    Attributes:
        max_retries: Maximum retry times of the request.
        allowed_retry_methods: The allowed methods for retrying request.
        allowed_retry_status: The allowed status for retrying request.
            If both methods and status are fitted, the retrying strategy will work.
        timeout: Timeout value of the request in seconds.
        is_internal: Whether the request is from internal.

    """

    def __init__(self) -> None:

        self.max_retries = 3
        self.allowed_retry_methods = ["HEAD", "OPTIONS", "POST", "PUT"]
        self.allowed_retry_status = [429, 500, 502, 503, 504]

        self.timeout = 30
        self.is_internal = False
        self._x_source = "PYTHON-SDK"


config = Config()


class TimeoutHTTPAdapter(HTTPAdapter):
    """This class defines the http adapter for setting the timeout value.

    Arguments:
        *args: Extra arguments to initialize TimeoutHTTPAdapter.
        timeout: Timeout value of the post request in seconds.
        **kwargs: Extra keyword arguments to initialize TimeoutHTTPAdapter.

    """

    def __init__(self, *args: Any, timeout: Optional[int] = None, **kwargs: Any) -> None:
        self.timeout = timeout if timeout is not None else config.timeout
        super().__init__(*args, **kwargs)

    def send(  # pylint: disable=too-many-arguments
        self,
        request: PreparedRequest,
        stream: Any = False,
        timeout: Any = None,
        verify: Any = True,
        cert: Any = None,
        proxies: Any = None,
    ) -> Any:
        """Send the request.

        Arguments:
            request: The PreparedRequest being sent.
            stream: Whether to stream the request content.
            timeout: Timeout value of the post request in seconds.
            verify: A path string to a CA bundle to use or
                a boolean which controls whether to verify the server's TLS certificate.
            cert: User-provided SSL certificate.
            proxies: Proxies dict applying to the request.

        Returns:
            Response object.

        """
        if timeout is None:
            timeout = self.timeout
        return super().send(request, stream, timeout, verify, cert, proxies)


class UserSession(Session):
    """This class defines UserSession."""

    def __init__(self) -> None:
        super().__init__()
        # self.session.hooks["response"] = [logging_hook]

        retry_strategy = Retry(
            total=config.max_retries,
            status_forcelist=config.allowed_retry_status,
            raise_on_status=False,
            **{_ALLOWED_METHODS: config.allowed_retry_methods},
        )

        self.mount("http://", TimeoutHTTPAdapter(20, 20, retry_strategy))
        self.mount("https://", TimeoutHTTPAdapter(20, 20, retry_strategy))

    def request(  # type: ignore[override]
        self, method: str, url: str, *args: Any, **kwargs: Any
    ) -> Response:  # noqa: DAR401
        """Make the request.

        Arguments:
            method: Method for the request.
            url: URL for the request.
            *args: Extra arguments to make the request.
            **kwargs: Extra keyword arguments to make the request.

        Returns:
            Response of the request.

        Raises:
            ResponseError: If post response error.

        """
        try:
            response = super().request(method, url, *args, **kwargs)
            if response.status_code not in (200, 201):
                logger.error(
                    "Unexpected status code(%d)!%s", response.status_code, ResponseLogging(response)
                )
                raise ResponseError(response=response)

            logger.debug(ResponseLogging(response))
            return response

        except RequestException as error:
            logger.error(
                "%s.%s: %s%s",
                error.__class__.__module__,
                error.__class__.__name__,
                error,
                RequestLogging(error.request),
            )
            raise


SESSIONS: DefaultDict[int, UserSession] = defaultdict(UserSession)


def get_session() -> UserSession:
    """Create and return a session per PID so each sub-processes will use their own session.

    Returns:
        The session corresponding to the process.
    """
    return SESSIONS[os.getpid()]


class UserResponse:
    """This class used to read data from Response with stream method.

    Arguments:
        response: Response of the Session.request().

    """

    def __init__(self, response: Response):
        self.response = response

    def __enter__(self) -> "UserResponse":
        return self

    def __exit__(self, *_: Any) -> None:
        self.close()

    def close(self) -> None:
        """Close the response."""
        self.response.close()

    def read(self, amt: Optional[int] = None) -> bytes:
        """Read data from response.

        Arguments:
            amt: The needed read amount.

        Returns:
            Response of the request.

        """
        if amt is None:
            try:
                return b"".join(chunk for chunk in self.response.iter_content(_CHUNK_SIZE))

            except StreamConsumedError:
                return b""

        try:
            return next(self.response.iter_content(amt))  # type: ignore[no-any-return]

        except StopIteration:
            return b""


_T = TypeVar("_T")


def submit_multithread_tasks(
    function: Callable[[_T], Any], arguments: Iterable[_T], *, jobs: int
) -> None:
    """Multi-thread framework.

    Arguments:
        function: The function to call.
        arguments: The arguments of the function.
        jobs: The number of the max workers in multi-thread call procession.

    """
    with ThreadPoolExecutor(jobs) as executor:
        futures = [executor.submit(function, argument) for argument in arguments]
        done, not_done = wait(futures, return_when=FIRST_EXCEPTION)

        for future in not_done:
            future.cancel()
        for future in done:
            future.result()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Dataframe module."""

from graviti.dataframe.column.series import ArraySeries, FileSeries
from graviti.dataframe.column.series import Series as ColumnSeries
from graviti.dataframe.container import RECORD_KEY, Container
from graviti.dataframe.frame import DataFrame

__all__ = ["ArraySeries", "ColumnSeries", "Container", "DataFrame", "FileSeries", "RECORD_KEY"]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Workspace."""

from graviti.manager import DatasetManager
from graviti.openapi import get_current_user
from graviti.utility import urlnorm


class Workspace:
    """This class defines the initial client to interact between local and server.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.

    """

    _DEFAULT_URL_CN = "https://api.graviti.cn"
    _DEFAULT_URL_COM = "https://api.graviti.com"

    def __init__(
        self,
        access_key: str,
        url: str = "",
    ) -> None:
        if access_key.startswith("Accesskey-"):
            url = url if url else Workspace._DEFAULT_URL_CN
        elif access_key.startswith("ACCESSKEY-"):
            url = url if url else Workspace._DEFAULT_URL_COM
        else:
            raise TypeError("Wrong accesskey format!")

        if not url.startswith("https://"):
            raise TypeError("Invalid url, only support url starts with 'https://'")

        url = urlnorm(url)

        self._access_key = access_key
        self._url = url
        response = get_current_user(access_key, url)
        self._dataset_manager = DatasetManager(access_key, url, response["workspace"])

    @property
    def access_key(self) -> str:
        """Return the access key of the user.

        Returns:
            The access key of the user.

        """
        return self._access_key

    @property
    def url(self) -> str:
        """Return the url of the graviti website.

        Returns:
            The url of the graviti website.

        """
        return self._url

    @property
    def datasets(self) -> DatasetManager:
        """Get class :class:`~graviti.manager.dataset.DatasetManager` instance.

        Returns:
            Required :class:`~graviti.manager.dataset.DatasetManager` instance.

        """
        return self._dataset_manager



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the user."""

from typing import Any, Dict

from graviti.openapi.requests import open_api_do


def get_current_user(access_key: str, url: str) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/current-user`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_current_user("ACCESSKEY-********", "https://api.graviti.com")
        {
            "id": "41438e9df9a82a194e1e76cc31c1d8d4",
            "nickname": "czh ual",
            "email": "********@graviti.com",
            "mobile": null,
            "description": "",
            "workspace": "graviti-example",
            "team": null
        }

    """
    url = f"{url}/v2/current-user"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Portex external base class."""


from inspect import Signature
from typing import TYPE_CHECKING, Any, ClassVar

import pyarrow as pa

from graviti.portex.base import PortexType
from graviti.portex.register import ExternalContainerRegister, ExternalElementResgister

if TYPE_CHECKING:
    from graviti.portex.builtin import PortexBuiltinType
    from graviti.portex.factory import TypeFactory
    from graviti.portex.package import ExternalPackage

EXTERNAL_TYPE_TO_CONTAINER = ExternalContainerRegister.EXTERNAL_TYPE_TO_CONTAINER
EXTERNAL_TYPE_TO_ELEMENT = ExternalElementResgister.EXTERNAL_TYPE_TO_ELEMENT


class PortexExternalType(PortexType):  # pylint: disable=abstract-method
    """The base class of Portex external type."""

    _signature: ClassVar[Signature]
    _factory: ClassVar["TypeFactory"]

    package: ClassVar["ExternalPackage"]

    def __init_subclass__(cls) -> None:
        cls._signature = cls.params.get_signature()

        key = (cls.package.url, cls.package.revision, cls.__name__)
        base_class = cls._factory.class_

        container = EXTERNAL_TYPE_TO_CONTAINER.get(key) or getattr(base_class, "container", None)
        if container:
            cls.container = container

        element = EXTERNAL_TYPE_TO_ELEMENT.get(key) or getattr(cls._factory.class_, "element", None)
        if element:
            cls.element = element

        search_container = getattr(base_class, "search_container", None)
        if search_container:
            cls.search_container = search_container

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        bound_arguments = self._signature.bind(*args, **kwargs)
        bound_arguments.apply_defaults()
        arguments = bound_arguments.arguments

        for key, value in arguments.items():
            arguments[key] = self.params[key].check(value)

        self.__dict__.update(arguments)

    def _get_column_count(self) -> int:
        """Get the total column count of the portex type.

        Returns:
            The total column count.

        """
        return self.to_builtin()._get_column_count()  # pylint: disable=protected-access

    @property
    def internal_type(self) -> PortexType:
        """Get the internal type of the PortexExternalType.

        Returns:
            The internal type of the PortexExternalType.

        """
        return self._factory({name: getattr(self, name) for name in self.params})

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return self.internal_type.to_pyarrow()

    def to_builtin(self) -> "PortexBuiltinType":
        """Expand the top level of the Portex external type to Portex builtin type.

        Returns:
            The expanded Portex builtin type.

        """
        return self.internal_type.to_builtin()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the search."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do


def create_search(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    commit_id: str,
    sheet: str,
    criteria: Dict[str, Any],
    search_id: Optional[str] = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{owner}/{dataset}/commits/{commit_id}/sheets\
    /{sheet}/search`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        commit_id: The commit id.
        sheet: The sheet name.
        criteria: The criteria of the search.
        search_id: The search id of this search.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is to get
            all search results.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> create_search(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "BDD100K",
        ...     commit_id = "fde63f357daf46088639e9f57fd81cad",
        ...     sheet = "train",
        ...     criteria = {
        ...         "where": {
        ...            "$or": [
        ...                {
        ...                    "$eq": ["$.filename", "0000f77c-6257be58.jpg"],
        ...                },
        ...                {
        ...                    "$and": [
        ...                        {
        ...                            "$eq": ["$.attribute.weather", "clear"]
        ...                        },
        ...                        {
        ...                            "$eq": ["$.attribute.timeofday", "daytime"]
        ...                        }
        ...                    ]
        ...                }
        ...            ]
        ...        },
        ...        "offset": 0,
        ...        "limit": 128
        ...    }
        ... )
        {
            "data": [
                {
                    "__record_key": "123750493121329585",
                    "filename": "0000f77c-6257be58.jpg",
                    "image": {
                        "url": "https://content-store-prod-vers",
                        "checksum": "dcc197970e607f7576d978972f6fb312911ce005"
                    },
                    "attribute": {
                        "weather": "clear",
                        "scene": "city street",
                        "timeofday": "daytime"
                    }
                },
                ...(total 128 items)
            ],
            "search_id": "5c7de503c88446e8b37a258f71783d7d"
        }

        Use search_id to avoid creating new search.

        >>> create_search(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "BDD100K",
        ...     commit_id = "fde63f357daf46088639e9f57fd81cad",
        ...     sheet = "train",
        ...     criteria = {"where": {"$eq": ["$.filename", "0000f77c-6257be58.jpg"]}}
        ...     search_id = "5c7de503c88446e8b37a258f71783d7d"
        ... )
        {
            "data": [
                {
                    "__record_key": "123750493121329585",
                    "filename": "0000f77c-6257be58.jpg",
                    "image": {
                        "url": "https://content-store-prod-vers",
                        "checksum": "dcc197970e607f7576d978972f6fb312911ce005"
                    },
                    "attribute": {
                        "weather": "clear",
                        "scene": "city street",
                        "timeofday": "daytime"
                    }
                },
                ...(total 128 items)
            ],
            "search_id": "5c7de503c88446e8b37a258f71783d7d"
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/commits/{commit_id}/sheets/{sheet}/search"

    criteria = criteria.copy()
    if offset is not None:
        criteria["offset"] = offset
    if limit is not None:
        criteria["limit"] = limit
    post_data: Dict[str, Any] = {"criteria": criteria}
    if search_id:
        post_data["search_id"] = search_id

    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Graviti audio file class."""

from graviti.file.base import File, RemoteFile
from graviti.portex import STANDARD_URL, ExternalElementResgister


class Audio(File):
    """This class represents local audio files."""

    __slots__ = File.__slots__


@ExternalElementResgister(STANDARD_URL, "main", "file.Audio")
class RemoteAudio(RemoteFile):
    """This class represents remote audio files."""

    __slots__ = RemoteFile.__slots__



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the data."""

from typing import Any, Dict, List, Optional, Tuple, Union

from graviti.openapi.requests import open_api_do


def _list_data(
    access_key: str,
    url: str,
    *,
    columns: Optional[str],
    order_by: Optional[str],
    offset: Optional[int],
    limit: Optional[int],
) -> Dict[str, Any]:

    params: Dict[str, Any] = {}

    if columns is not None:
        params["columns"] = columns
    if order_by is not None:
        params["order_by"] = order_by
    if offset is not None:
        params["offset"] = offset
    if limit is not None:
        params["limit"] = limit

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def list_draft_data(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
    columns: Optional[str] = None,
    order_by: Optional[str] = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/drafts/{draft_number}\
    /sheets/{sheet}/data`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        sheet: The sheet name.
        columns: The string of column names separated by ``|``.
            Multiple indexes can be expressed using ``.``. None means to get all columns.
        order_by: The string of column names separated by ``|`` whose order determines the
            precedence of the sort. The rest are sorted by `__record_key` first. Multiple
            indexes can be expressed using ``.``.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_draft_data(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     draft_number = 1,
        ...     sheet = "train",
        ...     order_by = "filename|attribute.weather",
        ... )
        {
            "data": [
                {
                    "__record_key": "123750493121329585",
                    "filename": "0000f77c-6257be58.jpg",
                    "image": {
                        "url": "https://content-store-prod-vers",
                        "checksum": "dcc197970e607f7576d978972f6fb312911ce005"
                    },
                    "attribute": {
                        "weather": "clear",
                        "scene": "city street",
                        "timeofday": "daytime"
                    },
                    "box2ds": [
                        {
                            "xmin": 1125.902264,
                            "xmax": 1156.978645,
                            "ymin": 133.184488,
                            "ymax": 210.875445,
                            "category": "traffic light",
                            "attribute": {
                                "occluded": false,
                                "truncated": false,
                                "trafficLightColor": "G"
                            }
                        },
                        {
                            "xmin": 1156.978645,
                            "xmax": 1191.50796,
                            "ymin": 136.637417,
                            "ymax": 210.875443,
                            "category": "traffic light",
                            "attribute": {
                                "occluded": false,
                                "truncated": false,
                                "trafficLightColor": "G"
                            }
                        },
                ...
                    ]
                },
                ...(total 128 items)
            ],
            "offset": 0,
            "record_size": 128,
            "total_count": 70000
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/drafts/{draft_number}/sheets/{sheet}/data"

    return _list_data(
        access_key, url, columns=columns, order_by=order_by, offset=offset, limit=limit
    )


def list_commit_data(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    commit_id: str,
    sheet: str,
    columns: Optional[str] = None,
    order_by: Optional[str] = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/commits/{commit_id}/sheets\
    /{sheet}/data`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        commit_id: The commit id.
        sheet: The sheet name.
        columns: The string of column names separated by ``|``.
            Multiple indexes can be expressed using ``.``. None means to get all columns.
        order_by: The string of column names separated by ``|`` whose order determines the
            precedence of the sort. The rest are sorted by `__record_key` first. Multiple
            indexes can be expressed using ``.``.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_commit_data(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     commit_id = "fde63f357daf46088639e9f57fd81cad",
        ...     sheet = "train",
        ...     order_by = "filename|attribute.weather",
        ... )
        {
            "data": [
                {
                    "__record_key": "123750493121329585",
                    "filename": "0000f77c-6257be58.jpg",
                    "image": {
                        "url": "https://content-store-prod-vers",
                        "checksum": "dcc197970e607f7576d978972f6fb312911ce005"
                    },
                    "attribute": {
                        "weather": "clear",
                        "scene": "city street",
                        "timeofday": "daytime"
                    },
                    "box2ds": [
                        {
                            "xmin": 1125.902264,
                            "xmax": 1156.978645,
                            "ymin": 133.184488,
                            "ymax": 210.875445,
                            "category": "traffic light",
                            "attribute": {
                                "occluded": false,
                                "truncated": false,
                                "trafficLightColor": "G"
                            }
                        },
                        {
                            "xmin": 1156.978645,
                            "xmax": 1191.50796,
                            "ymin": 136.637417,
                            "ymax": 210.875443,
                            "category": "traffic light",
                            "attribute": {
                                "occluded": false,
                                "truncated": false,
                                "trafficLightColor": "G"
                            }
                        },
                ...
                    ]
                },
                ...(total 128 items)
            ],
            "offset": 0,
            "record_size": 128,
            "total_count": 70000
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/commits/{commit_id}/sheets/{sheet}/data"

    return _list_data(
        access_key, url, columns=columns, order_by=order_by, offset=offset, limit=limit
    )


def update_data(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
    data: Union[List[Dict[str, Any]], Tuple[Dict[str, Any], ...]],
) -> None:
    """Execute the OpenAPI `PATCH /v2/datasets/{owner}/{dataset}/drafts/{draft_number}\
    /sheets/{sheet}/data`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        sheet: The sheet name.
        data: The update data.

    Examples:
        >>> update_data(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "OxfordIIITPet",
        ...     draft_number = 1,
        ...     sheet = "train",
        ...     data = [
        ...         {
        ...             "__record_key": "123750493121329585",
        ...             "filename": "0000f77c-6257be58.jpg",
        ...             "image": {
        ...                 "checksum": "dcc197970e607f7576d978972f6fb312911ce005"
        ...             },
        ...             "attribute": {
        ...                 "weather": "clear",
        ...                 "scene": "city street",
        ...                 "timeofday": "daytime"
        ...             },
        ...         },
        ...         {
        ...             "__record_key": "123750493121329585",
        ...             "filename": "0000f77c-62c2a288.jpg",
        ...             "image": {
        ...                 "checksum": "dcc197970e607f7576d978972f6fb2a2881ce004"
        ...             },
        ...             "attribute": {
        ...                 "weather": "clear",
        ...                 "scene": "highway",
        ...                 "timeofday": "dawn/dusk"
        ...             },
        ...         }
        ...     ],
        ... )

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/drafts/{draft_number}/sheets/{sheet}/data"
    patch_data = {"data": data}

    open_api_do("PATCH", access_key, url, json=patch_data)


def add_data(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
    data: Union[List[Dict[str, Any]], Tuple[Dict[str, Any], ...]],
    strategy_arguments: Optional[Dict[str, Any]] = None,
) -> None:
    """Execute the OpenAPI `POST /v2/datasets/{owner}/{dataset}/drafts/{draft_number}\
    /sheets/{sheet}/data`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        sheet: The sheet name.
        data: The update data.
        strategy_arguments: Arguments required by the ``__record_key`` generation strategy
            of the sheet.

    Examples:
        >>> add_data(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "OxfordIIITPet",
        ...     draft_number = 1,
        ...     sheet = "train",
        ...     data = [
        ...         {
        ...             "filename": "0000f77c-6257be58.jpg",
        ...             "image": {
        ...                 "checksum": "dcc197970e607f7576d978972f6fb312911ce005"
        ...             },
        ...             "attribute": {
        ...                 "weather": "clear",
        ...                 "scene": "city street",
        ...                 "timeofday": "daytime"
        ...             },
        ...         },
        ...         {
        ...             "filename": "0000f77c-62c2a288.jpg",
        ...             "image": {
        ...                 "checksum": "dcc197970e607f7576d978972f6fb2a2881ce004"
        ...             },
        ...             "attribute": {
        ...                 "weather": "clear",
        ...                 "scene": "highway",
        ...                 "timeofday": "dawn/dusk"
        ...             },
        ...         }
        ...     ],
        ... )

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/drafts/{draft_number}/sheets/{sheet}/data"
    post_data: Dict[str, Any] = {"data": data}
    if strategy_arguments is not None:
        post_data["strategy_arguments"] = strategy_arguments

    open_api_do("POST", access_key, url, json=post_data)


def delete_data(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
    record_keys: List[str],
) -> None:
    """Execute the OpenAPI `DELETE /v2/datasets/{owner}/{dataset}/drafts/{draft_number}\
    /sheets/{sheet}/data`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        sheet: The sheet name.
        record_keys: The record keys of the data to be deleted.

    Examples:
        >>> delete_data(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "OxfordIIITPet",
        ...     draft_number = 1,
        ...     sheet = "train",
        ...     record_keys = ["123750493121329585", "123750493121329586"],
        ... )

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/drafts/{draft_number}/sheets/{sheet}/data"
    open_api_do("DELETE", access_key, url, json={"record_keys": record_keys})



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Repr related methods."""

from abc import ABC
from datetime import datetime
from enum import Enum, auto
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Mapping,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

MAX_REPR_ROWS = 10
INDENT = " " * 2


class ReprType(Enum):
    """ReprType is an enumeration type.

    It defines the repr strategy type and includes 'INSTANCE', 'SEQUENCE' and 'MAPPING'.

    """

    INSTANCE = auto()
    SEQUENCE = auto()
    MAPPING = auto()


class ReprMixin:
    """ReprMixin provides customized repr config and method."""

    __slots__ = ()

    _repr_type: ClassVar[ReprType] = ReprType.INSTANCE
    _repr_attrs: Iterable[str] = ()
    _repr_maxlevel = 1
    _repr_non_empty = False

    def __repr__(self) -> str:
        return _repr1(self, self._repr_maxlevel, self._repr_maxlevel, True)

    def __str__(self) -> str:
        return _repr1(self, self._repr_maxlevel, self._repr_maxlevel, False)

    def _repr_head(self) -> str:
        return self.__class__.__name__


class _ReprConfig:
    """Config for customized repr method.

    Arguments:
        indent: The indent width for different level repr string.
        maxlist: The max items displayed in the sequence repr string.
        maxdict: The max items displayed in the mapping repr string.

    """

    def __init__(self, indent: int = 2, maxlist: int = 16, maxdict: int = 16) -> None:
        self.indent = indent
        self.maxlist = maxlist
        self.maxdict = maxdict

    @property
    def indent(self) -> int:
        """The getter of repr indent width.

        Returns:
            The indent width of repr config.

        """
        return self._indent

    @indent.setter
    def indent(self, indent: int) -> None:
        """The setter of repr indent width.

        Arguments:
            indent: The width of the ident to be set.

        """
        self._indent = indent
        self._indent_str = " " * indent


repr_config = _ReprConfig()


# Explicitly inherit from 'abc.ABC' to tell pylint it is a abstract class
# Reference:
# - https://github.com/PyCQA/pylint/issues/3098
# - https://github.com/PyCQA/pylint/pull/3446
class _ReprSequence(ReprMixin, Sequence[Any], ABC):
    ...


class _ReprMapping(ReprMixin, Mapping[Any, Any], ABC):
    ...


_ReprPrinter = Callable[[Any, int, int, bool], str]

_PRINTERS: Dict[Union[Type[Any], ReprType], _ReprPrinter] = {}


class _PrinterRegister:
    """Decorator class to register repr printer functions to '_PRINTERS'.

    Arguments:
        key: The key of the '_PRINTERS' for repr dispatch.

    """

    _S = TypeVar("_S", bound=_ReprPrinter)

    def __init__(self, key: Union[Type[Any], ReprType]) -> None:
        self._key = key

    def __call__(self, printer: _S) -> _S:
        _PRINTERS[self._key] = printer
        return printer


def _repr1(obj: Any, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    class_: Type[Any] = type(obj)
    printer = _PRINTERS.get(getattr(class_, "_repr_type", class_), None)
    return printer(obj, level, maxlevel, folding) if printer else repr(obj)


@_PrinterRegister(ReprType.INSTANCE)
def _repr_instance(obj: ReprMixin, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for `ReprType.INSTANCE`.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    # pylint: disable=protected-access
    return f"{obj._repr_head()}{_repr_attrs(obj, level, maxlevel, folding)}"


@_PrinterRegister(ReprType.SEQUENCE)
def _repr_sequence(obj: _ReprSequence, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for `ReprType.SEQUENCE`.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    # pylint: disable=protected-access
    return (
        f"{obj._repr_head()} {_repr_builtin_list(obj, level, maxlevel, folding)}"
        f"{_repr_attrs(obj, level, maxlevel, folding)}"
    )


@_PrinterRegister(ReprType.MAPPING)
def _repr_mapping(obj: _ReprMapping, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for `ReprType.MAPPING`.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    # pylint: disable=protected-access
    return (
        f"{obj._repr_head()} {_repr_builtin_dict(obj, level, maxlevel, folding)}"
        f"{_repr_attrs(obj, level, maxlevel, folding)}"
    )


@_PrinterRegister(list)
def _repr_builtin_list(obj: Sequence[Any], level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for buildin type list.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    return _repr_builtin_sequence(obj, level, maxlevel, folding, "[", "]")


@_PrinterRegister(tuple)
def _repr_builtin_tuple(obj: Sequence[Any], level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for buildin type tuple.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    return _repr_builtin_sequence(obj, level, maxlevel, folding, "(", ")")


@_PrinterRegister(dict)
def _repr_builtin_dict(obj: Mapping[Any, Any], level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for buildin type dict.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    if getattr(obj, "_repr_non_empty", False) and level <= 0:
        return "{...}"

    object_length = len(obj)
    if object_length == 0:
        return "{}"
    if level <= 0:
        return "{...}"
    newlevel = level - 1

    keys = tuple(obj)

    fold, unfold = _calculate_fold_number(object_length, repr_config.maxdict, folding)
    pieces = [
        f"{repr(key)}: {_repr1(obj[key], newlevel, maxlevel, folding)}" for key in keys[:unfold]
    ]
    if fold:
        key = keys[-1]
        pieces.append(f"... ({fold} items are folded)")
        pieces.append(f"{repr(key)}: {_repr1(obj[key], newlevel, maxlevel, folding)}")

    return _join_with_indent(pieces, level, maxlevel, "{", "}")


@_PrinterRegister(datetime)
def _repr_datetime(obj: datetime, _: int, __: int, ___: bool) -> str:
    return str(obj)


def _repr_attrs(obj: Any, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for object attributes.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    # pylint: disable=protected-access
    if not obj._repr_attrs or level <= 0:
        return ""

    attributes = []
    for attribute in obj._repr_attrs:
        value = getattr(obj, attribute, ...)
        if value is not ...:
            attributes.append((attribute, value))

    if not attributes:
        return "()"

    newlevel = level - 1
    pieces = (f"({key}): {_repr1(value, newlevel, maxlevel, folding)}" for key, value in attributes)

    return _join_with_indent(pieces, level, maxlevel, "(", ")")


def _repr_builtin_sequence(  # pylint: disable=too-many-arguments
    obj: Sequence[Any], level: int, maxlevel: int, folding: bool, left: str, right: str
) -> str:
    """Customized repr method for sequence.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.
        left: The left bracket symbol.
        right: The right bracket symbol.

    Returns:
        "repr" of the object.

    """
    if getattr(obj, "_repr_non_empty", False) and level <= 0:
        return f"{left}...{right}"

    object_length = len(obj)
    if object_length == 0:
        return f"{left}{right}"
    if level <= 0:
        return f"{left}...{right}"
    newlevel = level - 1

    fold, unfold = _calculate_fold_number(object_length, repr_config.maxlist, folding)
    pieces = [_repr1(item, newlevel, maxlevel, folding) for item in obj[:unfold]]
    if fold:
        pieces.append(f"... ({fold} items are folded)")
        pieces.append(_repr1(obj[-1], newlevel, maxlevel, folding))

    return _join_with_indent(pieces, level, maxlevel, left, right)


def _calculate_fold_number(length: int, max_length: int, folding: bool) -> Tuple[int, int]:
    """Calculate the number of the folded and unfolded items.

    Arguments:
        length: The object length.
        max_length: The configured max length.
        folding: Whether fold the "repr" of the object.

    Returns:
        The number of the folded items & the number of the unfolded items.

    """
    if folding and length > max_length:
        return length - max_length + 1, max_length - 2
    return 0, length


def _join_with_indent(
    pieces: Iterable[str], level: int, maxlevel: int, left: str, right: str
) -> str:
    """Handle the indent, newline and "," of the string list.

    Arguments:
        pieces: The string list need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        left: The left bracket symbol.
        right: The right bracket symbol.

    Returns:
        Repr string with indent, newline and ",".

    """
    # pylint: disable=protected-access
    inner_indent = repr_config._indent_str * (maxlevel - level + 1)
    outer_indent = repr_config._indent_str * (maxlevel - level)

    sep = f",\n{inner_indent}"
    return f"{left}\n{inner_indent}{sep.join(pieces)}\n{outer_indent}{right}"



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Common tools."""

import base64
import json
from typing import Tuple

import pyarrow as pa

from graviti.portex import convert_portex_schema_to_avro, record


def get_schema(schema: record) -> Tuple[str, str, str]:
    """Get portex schema, avro schema and arrow schema.

    Arguments:
        schema: The portex schema.

    Returns:
        The tuple of portex schema, avro schema and arrow schema.
    """
    portex_schema = schema.to_yaml()
    avro_schema = json.dumps(convert_portex_schema_to_avro(schema))
    pyarrow_schema = pa.schema(schema.fields.to_pyarrow())
    arrow_schema = base64.encodebytes(pyarrow_schema.serialize().to_pybytes()).decode("ascii")

    return portex_schema, avro_schema, arrow_schema



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Sheets."""

from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    ItemsView,
    Iterator,
    KeysView,
    List,
    MutableMapping,
    ValuesView,
)

import pyarrow as pa
from tqdm.auto import tqdm

import graviti.portex as pt
from graviti.dataframe import RECORD_KEY, DataFrame
from graviti.exception import FieldNameConflictError
from graviti.manager.common import LIMIT
from graviti.operation import AddData, CreateSheet, DeleteSheet, SheetOperation
from graviti.paging import LazyLowerCaseFactory
from graviti.portex import PortexRecordBase
from graviti.utility import ReprMixin

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset


class Sheets(MutableMapping[str, DataFrame], ReprMixin):
    """The basic structure of the Graviti sheets."""

    _data: Dict[str, DataFrame]
    _dataset: "Dataset"
    operations: List[SheetOperation] = []

    def __len__(self) -> int:
        return self._get_data().__len__()

    def __getitem__(self, key: str) -> DataFrame:
        return self._get_data().__getitem__(key)

    def __setitem__(self, key: str, value: DataFrame) -> None:
        if value.operations is not None:
            raise NotImplementedError(
                "Not support assigning one DataFrame to multiple sheets."
                " Please use method 'copy' first."
            )
        is_replace = False
        if key in self:
            is_replace = True

        self._get_data().__setitem__(key, value)
        if is_replace:
            self.operations.append(DeleteSheet(key))

        value.operations = [AddData(value.copy())]

        assert isinstance(value.schema, pt.record)
        self.operations.append(CreateSheet(key, value.schema.copy()))

    def __delitem__(self, key: str) -> None:
        df = self[key]
        self._get_data().__delitem__(key)
        del df.operations
        self.operations.append(DeleteSheet(key))

    def __iter__(self) -> Iterator[str]:
        return self._get_data().__iter__()

    def _list_data(self, offset: int, limit: int, sheet_name: str) -> Dict[str, Any]:
        raise NotImplementedError

    def _list_sheets(self) -> Dict[str, Any]:
        raise NotImplementedError

    def _get_sheet(self, sheet_name: str) -> Dict[str, Any]:
        raise NotImplementedError

    def _init_dataframe(self, sheet_name: str) -> DataFrame:
        sheet = self._get_sheet(sheet_name)
        schema = PortexRecordBase.from_yaml(sheet["schema"])

        patype = schema.to_pyarrow()

        factory = LazyLowerCaseFactory(
            sheet["record_count"],
            LIMIT,
            partial(self._list_data, sheet_name=sheet_name),
            pa.struct([pa.field(RECORD_KEY, pa.string()), *patype]),
        )
        df = DataFrame._from_factory(  # pylint: disable=protected-access
            factory, schema, object_permission_manager=self._dataset.object_permission_manager
        )
        df.operations = []

        return df

    def _init_data(self) -> None:
        self._data = {}
        sheets_info = self._list_sheets()["sheets"]

        for sheet_info in sheets_info:
            sheet_name = sheet_info["name"]
            self._data[sheet_name] = self._init_dataframe(sheet_name)

    def _get_data(self) -> Dict[str, DataFrame]:
        if not hasattr(self, "_data"):
            self._init_data()

        return self._data

    def _check_record_names(self, schema: PortexRecordBase, sheet_name: str) -> None:
        name_mapping: Dict[str, str] = {}
        for name, portex_type in schema.items():
            builtin_type = portex_type.to_builtin()
            if isinstance(builtin_type, pt.record):
                self._check_record_names(builtin_type, sheet_name)
            elif isinstance(builtin_type, pt.array):
                self._check_array_names(builtin_type, sheet_name)

            lower_name = name.lower()
            if lower_name in name_mapping:
                raise FieldNameConflictError(
                    f'In sheet "{sheet_name}", the field name "{name}" is conflict with '
                    f'"{name_mapping[lower_name]}", the Graviti data platform is case insensitive.'
                )

            name_mapping[lower_name] = name

    def _check_array_names(self, schema: pt.array, sheet_name: str) -> None:
        items = schema.items
        builtin_type = items.to_builtin()

        if isinstance(builtin_type, pt.record):
            self._check_record_names(builtin_type, sheet_name)
        elif isinstance(builtin_type, pt.array):
            self._check_array_names(builtin_type, sheet_name)

    def _upload_to_draft(self, draft_number: int, jobs: int, quiet: bool) -> None:
        """Upload the local dataset to Graviti.

        Arguments:
            draft_number: The number of the draft.
            jobs: The number of the max workers in multi-thread upload, the default is 8.
            quiet: Set to True to stop showing the upload process bar.

        """
        for sheet_name in {
            operation.sheet
            for operation in self.operations
            if not isinstance(operation, DeleteSheet)
        }:
            self._check_record_names(self[sheet_name].schema, sheet_name)

        dataset = self._dataset
        object_permission_manager = dataset.object_permission_manager
        for sheet_operation in self.operations:
            sheet_operation.do(
                dataset.access_key,
                dataset.url,
                dataset.owner,
                dataset.name,
                draft_number=draft_number,
            )
        self.operations = []

        df_total = 0
        file_total = 0
        for df in self.values():
            for df_operation in df.operations:  # type: ignore[union-attr]
                df_total += df_operation.get_data_count()
                file_total += df_operation.get_file_count()

        # Note that after done uploading, the two process bars will switch position due to the tqdm
        # bug https://github.com/tqdm/tqdm/issues/1000.
        with tqdm(
            total=file_total, disable=(file_total == 0 or quiet), desc="uploading binary files"
        ) as file_pbar:
            with tqdm(total=df_total, disable=quiet, desc="uploading structured data") as data_pbar:
                for sheet_name, df in self.items():
                    if not df.operations:
                        continue

                    for df_operation in df.operations:
                        df_operation.do(
                            dataset.access_key,
                            dataset.url,
                            dataset.owner,
                            dataset.name,
                            draft_number=draft_number,
                            sheet=sheet_name,
                            jobs=jobs,
                            data_pbar=data_pbar,
                            file_pbar=file_pbar,
                            object_permission_manager=object_permission_manager,
                        )
                    df.operations = []

    def keys(self) -> KeysView[str]:
        """Return a new view of the keys in sheets.

        Returns:
            The keys in dict.

        """
        return self._get_data().keys()

    def values(self) -> ValuesView[DataFrame]:
        """Return a new view of the values in sheets.

        Returns:
            The values in dict.

        """
        return self._get_data().values()

    def items(self) -> ItemsView[str, DataFrame]:
        """Return a new view of the (key, value) pairs in sheets.

        Returns:
            The (key, value) pairs in dict.

        """
        return self._get_data().items()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Portex enum values releated classes."""


from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, Union

from graviti.utility import INDENT, UserMapping, UserSequence

_Types = (str, int, bool, type(None))
EnumValueType = Union[str, int, bool, None]


class EnumValues:
    """The base class of portex enum values."""

    index_scope: Tuple[int, int]
    value_to_index: Dict[EnumValueType, Optional[int]]
    index_to_value: Dict[Optional[int], EnumValueType]

    def __repr__(self) -> str:
        return self._repr1(0)

    def _init_mapping(self, value_to_index: Dict[EnumValueType, Optional[int]]) -> None:
        if None not in value_to_index:
            value_to_index[None] = None

        self.value_to_index = value_to_index
        self.index_to_value = {value: key for key, value in value_to_index.items()}

    @staticmethod
    def _check_value_type(value: Any) -> EnumValueType:
        if not isinstance(value, _Types):
            raise TypeError(
                f"The type of enum value ({value}) must be 'str', 'int', 'bool' or None"
            )

        return value

    def _repr1(self, level: int) -> str:
        raise NotImplementedError

    def to_pyobj(self) -> Union[List[EnumValueType], Dict[int, EnumValueType]]:
        """Dump the instance to a python list or dict.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class EnumValueList(EnumValues, UserSequence[EnumValueType]):
    """The portex enum values in list format.

    Arguments:
        values: The enum values.

    """

    _data: List[EnumValueType]

    def __init__(self, values: Iterable[EnumValueType]) -> None:
        checker = self._check_value_type
        self._data = [checker(value) for value in values]
        self._init()

    def _init(self) -> None:
        self.index_scope = (0, len(self._data) - 1)

        value_to_index: Dict[EnumValueType, Optional[int]] = {
            value: key for key, value in enumerate(self._data)
        }
        self._init_mapping(value_to_index)

    def _repr1(self, level: int) -> str:
        lines = ["["]
        lines.extend(f"{INDENT}{repr(value)}," for value in self._data)
        lines.append("]")

        return f"\n{level * INDENT}".join(lines)

    def to_pyobj(self) -> List[EnumValueType]:
        """Dump the instance to a python list.

        Returns:
            A python list representation of the enum values.

        """
        return self._data.copy()


class EnumValueDict(EnumValues, UserMapping[int, EnumValueType]):
    """The portex enum values in dict format.

    Arguments:
        values: The enum values.

    """

    _data: Dict[int, EnumValueType]

    def __init__(self, values: Mapping[int, EnumValueType]) -> None:
        checker = self._check_value_type
        try:
            self._data = {int(key): checker(value) for key, value in values.items()}
        except ValueError as error:
            raise TypeError("The portex enum index type must be 'int'") from error

        self._init()

    def _init(self) -> None:
        value_to_index = {value: key for key, value in self._data.items()}

        self.index_scope = (min(value_to_index.values()), max(value_to_index.values()))

        self._init_mapping(value_to_index)  # type: ignore[arg-type]

    def _repr1(self, level: int) -> str:
        lines = ["{"]
        lines.extend(f"{INDENT}{key}: {repr(value)}," for key, value in self._data.items())
        lines.append("}")

        return f"\n{level * INDENT}".join(lines)

    def to_pyobj(self) -> Dict[int, EnumValueType]:
        """Dump the instance to a python dict.

        Returns:
            A python dict representation of the enum values.

        """
        return self._data.copy()


def create_enum_values(
    values: Union[Sequence[EnumValueType], Mapping[int, EnumValueType]]
) -> EnumValues:
    """The factory function of EnumValues.

    Arguments:
        values: The enum values.

    Returns:
        The EnumValues instance created by the input enum values.

    Raises:
        TypeError: When the input enum values is not in list or dict format.

    """
    if isinstance(values, Sequence):
        return EnumValueList(values)

    if isinstance(values, Mapping):
        return EnumValueDict(values)

    raise TypeError("portex enum values should be a list or a dict")



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the branch."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do


def create_branch(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    name: str,
    revision: str,
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{owner}/{dataset}/branches`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        name: The name of the branch.
        revision: The information to locate the specific commit, which can be the commit id,
            the branch name, or the tag name.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> create_branch(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     name="branch-1",
        ...     revision="main"
        ... )
        {
            "name": "main",
            "commit_id": "fde63f357daf46088639e9f57fd81cad",
            "parent_commit_id": "f68b1375454f459b8a486b8d1f4d9ddb",
            "title": "first commit",
            "description": "desc",
            "committer": "graviti-example",
            "committed_at": "2021-03-03T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/branches"
    post_data = {"name": name, "revision": revision}
    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()


def list_branches(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/branches`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_branches(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST"
        ... )
        {
           "branches": [
               {
                   "name": "main",
                   "commit_id": "fde63f357daf46088639e9f57fd81cad",
                   "parent_commit_id": "f68b1375454f459b8a486b8d1f4d9ddb",
                   "title": "first commit",
                   "description": "desc",
                   "committer": "graviti-example",
                   "committed_at": "2021-03-03T18:58:10Z"
               }
           ],
           "offset": 0,
           "record_size": 1,
           "total_count": 1
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/branches"

    params = {}
    if offset is not None:
        params["offset"] = offset
    if limit is not None:
        params["limit"] = limit

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def get_branch(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    branch: str,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/branches/{branch}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        branch: The name of the branch.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_branch(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     branch="main",
        ... )
        {
            "name": "main",
            "commit_id": "fde63f357daf46088639e9f57fd81cad",
            "parent_commit_id": "f68b1375454f459b8a486b8d1f4d9ddb",
            "title": "first commit",
            "description": "desc",
            "committer": "graviti-example",
            "committed_at": "2021-03-03T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/branches/{branch}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def delete_branch(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    branch: str,
) -> None:
    """Execute the OpenAPI `DELETE /v2/datasets/{owner}/{dataset}/branches/{branch}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        branch: The name of the branch.

    Examples:
        >>> delete_branch(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     branch="branch-1",
        ... )

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/branches/{branch}"
    open_api_do("DELETE", access_key, url)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the commit."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do


def commit_draft(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    title: str,
    description: Optional[str] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{owner}/{dataset}/commits`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        title: The draft title.
        description: The draft description.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> commit_draft(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     draft_number=2,
        ...     title="commit-2",
        ... )
        {
           "commit_id": "85c57a7f03804ccc906632248dc8c359",
           "parent_commit_id": "784ba0d3bf0a41f6a7bfd771d8c00fcb",
           "title": "upload data",
           "description": "",
           "committer": "graviti-example",
           "committed_at": "2021-03-03T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/commits"
    post_data: Dict[str, Any] = {"draft_number": draft_number, "title": title}

    if description:
        post_data["description"] = description

    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data
    ).json()


def list_commits(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    revision: Optional[str] = None,
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/commits`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        revision: The information to locate the specific commit, which can be the commit id,
            the branch name, or the tag name.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> list_commits(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ... )
        {
           "commits": [
               {
                   "commit_id": "85c57a7f03804ccc906632248dc8c359",
                   "parent_commitId": "784ba0d3bf0a41f6a7bfd771d8c00fcb",
                   "title": "upload data",
                   "description": "",
                   "committer": "graviti-example",
                   "committed_at": "2021-03-03T18:58:10Z"
               }
           ],
           "offset": 0,
           "record_size": 1,
           "total_count": 1
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/commits"

    params: Dict[str, Any] = {}
    if offset is not None:
        params["offset"] = offset
    if limit is not None:
        params["limit"] = limit
    if revision is not None:
        params["revision"] = revision

    return open_api_do("GET", access_key, url, params=params).json()  # type: ignore[no-any-return]


def get_commit(
    access_key: str, url: str, owner: str, dataset: str, *, commit_id: str
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/commits/{commit_id}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        commit_id: The commit ID.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_commit(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     commit_id="85c57a7f03804ccc906632248dc8c359"
        ... )
        {
            "commit_id": "85c57a7f03804ccc906632248dc8c359",
            "parent_commit_id": "784ba0d3bf0a41f6a7bfd771d8c00fcb",
            "title": "upload data",
            "description": "",
            "committer": "graviti-example",
            "committed_at": "2021-03-03T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/commits/{commit_id}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]


def get_revision(
    access_key: str, url: str, owner: str, dataset: str, *, revision: str
) -> Dict[str, Any]:
    """Execute the OpenAPI `GET /v2/datasets/{owner}/{dataset}/revisions/{revision}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        revision: The information to locate the specific commit, which can be the commit id,
            the branch name, or the tag name.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> get_revision(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "MNIST",
        ...     revision="branch-1"
        ... )
        {
           "type": "BRANCH",
           "commit_id": "85c57a7f03804ccc906632248dc8c359",
           "parent_commit_id": "784ba0d3bf0a41f6a7bfd771d8c00fcb",
           "title": "upload data",
           "description": "",
           "committer": "graviti-example",
           "committed_at": "2021-03-03T18:58:10Z"
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/revisions/{revision}"
    return open_api_do("GET", access_key, url).json()  # type: ignore[no-any-return]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the schema."""

from typing import Any, Dict, List, Optional

from graviti.openapi.requests import open_api_do


def update_schema(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    draft_number: int,
    sheet: str,
    patch: Optional[List[Dict[str, Any]]] = None,
    _schema: str,
    _avro_schema: str,
    _arrow_schema: Optional[str] = None,
) -> None:
    """Execute the OpenAPI `PATCH /v2/datasets/{owner}/{dataset}/drafts/{draft_number}\
    /sheets/{sheet}/schema`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        draft_number: The draft number.
        sheet: The sheet name.
        patch: The list of patch operations which describe the changes to the json portex schema.

    Examples:
        >>> update_schema(
        ...     "ACCESSKEY-********",
        ...     "https://api.graviti.com",
        ...     "graviti-example",
        ...     "MNIST",
        ...     draft_number = 1,
        ...     sheet = "train",
        ...     _schema = '{"imports": [{"repo": "https://github.com/Project-OpenBytes/portex-\
standard@main", "types": [{"name": "file.Image"}]}], "type": "record", "fields": [{"name": \
"filename", "type": "string"}, {"name": "image", "type": "file.Image"}]}',
        ...     _avro_schema = '{"type": "record", "name": "root", "namespace": "cn.graviti.portex"\
, "aliases": [], "fields": [{"name": "filename", "type": "string"}, {"name": "image", "type": \
{"type": "record", "name": "image", "namespace": "cn.graviti.portex.root", "aliases": [], \
"fields": [{"name": "checksum", "type": [null, "string"]}]}}]}',
        ... )

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/drafts/{draft_number}/sheets/{sheet}/schema"
    patch_data: Dict[str, Any] = {"_schema": _schema, "_avro_schema": _avro_schema}

    if _arrow_schema is not None:
        patch_data["_arrow_schema"] = _arrow_schema
    if patch is not None:
        patch_data["patch"] = patch

    open_api_do("PATCH", access_key, url, json=patch_data)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The basic concepts and methods of the Graviti OpenAPI."""

from typing import Any
from uuid import uuid4

from requests.models import Response

from graviti.__version__ import __version__
from graviti.exception import ResponseError, ResponseErrorRegister
from graviti.utility import config, get_session

RESPONSE_ERROR_DISTRIBUTOR = ResponseErrorRegister.RESPONSE_ERROR_DISTRIBUTOR


def do(method: str, url: str, **kwargs: Any) -> Response:  # pylint: disable=invalid-name
    """Send a request.

    Arguments:
        method: The method of the request.
        url: The URL of the request.
        **kwargs: Extra keyword arguments to send in the GET request.

    Returns:
        Response of the request.

    """
    return get_session().request(method=method, url=url, **kwargs)


def open_api_do(method: str, access_key: str, url: str, **kwargs: Any) -> Response:
    """Send a request to the Graviti OpenAPI.

    Arguments:
        method: The method of the request.
        access_key: User's access key.
        url: The URL of the graviti website.
        **kwargs: Extra keyword arguments to send in the POST request.

    Raises:
        ResponseError: When the status code OpenAPI returns is unexpected.

    Returns:
        Response of the request.

    """
    headers = kwargs.setdefault("headers", {})
    headers["X-Token"] = access_key
    headers["X-Source"] = f"{config._x_source}/{__version__}"  # pylint: disable=protected-access
    headers["X-Request-Id"] = uuid4().hex

    try:
        return do(method=method, url=url, **kwargs)
    except ResponseError as error:
        response = error.response
        raise RESPONSE_ERROR_DISTRIBUTOR.get(
            (response.status_code, response.json().get("code")), ResponseError
        )(response=response) from None



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""The base elements of Portex type."""


import json
from copy import deepcopy
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Type, TypeVar

import pyarrow as pa
import yaml

from graviti.portex.package import Imports, Package
from graviti.portex.register import PyArrowConversionRegister
from graviti.utility import INDENT, PathLike, UserMutableMapping

if TYPE_CHECKING:
    from graviti.dataframe import Container
    from graviti.dataframe.sql.container import ArrayContainer
    from graviti.portex.builtin import PortexBuiltinType
    from graviti.portex.factory import ConnectedFieldsFactory
    from graviti.portex.field import ConnectedFields
    from graviti.portex.param import Params

PYARROW_TYPE_ID_TO_PORTEX_TYPE = PyArrowConversionRegister.PYARROW_TYPE_ID_TO_PORTEX_TYPE

_T = TypeVar("_T", bound="PortexType")


class PortexType:
    """The base class of portex type."""

    nullable: bool
    package: ClassVar[Package[Any]]
    params: ClassVar["Params"]
    container: ClassVar[Type["Container"]]
    element: ClassVar[Type[Any]]
    search_container: ClassVar[Type["ArrayContainer"]]

    def __repr__(self) -> str:
        return self._repr1(0)

    @classmethod
    def _from_pyarrow(cls: Type[_T], pyarrow_type: pa.DataType) -> _T:
        raise NotImplementedError

    def _repr1(self, level: int) -> str:
        with_params = False
        indent = level * INDENT
        lines = [f"{self.__class__.__name__}("]
        for name, parameter in self.params.items():
            attr = getattr(self, name)
            if attr != parameter.default:
                with_params = True
                lines.append(
                    f"{INDENT}{name}="  # pylint: disable=protected-access
                    f"{attr._repr1(level + 1) if hasattr(attr, '_repr1') else repr(attr)},"
                )

        if with_params:
            lines.append(")")
            return f"\n{indent}".join(lines)

        return f"{lines[0]})"

    def _get_column_count(self) -> int:  # pylint: disable=no-self-use
        """Get the total column count of the portex type.

        Returns:
            The total column count.

        """
        return 1

    @property
    def imports(self) -> Imports:
        """Get the PortexType imports.

        Returns:
            The :class:`Imports` instance of this PortexType.

        """
        imports = Imports()

        cls = self.__class__
        imports[cls.__name__] = cls

        for name in self.params:
            argument = getattr(self, name)
            argument_imports = getattr(argument, "imports", None)
            if argument_imports:
                imports.update(argument_imports)

        return imports

    @classmethod
    def from_pyobj(
        cls: Type[_T], content: Dict[str, Any], _imports: Optional["Imports"] = None
    ) -> _T:
        """Create Portex type instance from python dict.

        Arguments:
            content: A python dict representing a Portex type.

        Returns:
            A Portex type instance created from the input python dict.

        """
        if _imports is None:
            _imports = Imports.from_pyobj(content.get("imports", []))

        class_: Type[_T] = _imports[content["type"]]  # type: ignore[assignment]

        assert issubclass(class_, cls)
        kwargs = {}
        for name, param in class_.params.items():
            kwarg = content.get(name, ...)
            if kwarg is not ...:
                kwargs[name] = param.load(kwarg, _imports)

        type_ = class_(**kwargs)  # type: ignore[call-arg]
        return type_

    @classmethod
    def from_pyarrow(cls: Type[_T], pyarrow_type: pa.DataType) -> _T:
        """Create Portex type instance from PyArrow type.

        Arguments:
            pyarrow_type: The PyArrow type.

        Raises:
            TypeError: When the PyArrow type is not supported.

        Returns:
            The created Portex type instance.

        """
        try:
            portex_type = PYARROW_TYPE_ID_TO_PORTEX_TYPE[pyarrow_type.id]
        except KeyError:
            raise TypeError(f'Not supported PyArrow type "{pyarrow_type}"') from None

        # pylint: disable=protected-access
        return portex_type._from_pyarrow(pyarrow_type)  # type: ignore[return-value]

    @classmethod
    def from_json(cls: Type[_T], content: str) -> _T:
        """Create Portex type instance from JSON string.

        Arguments:
            content: A JSON string representing a Portex type.

        Returns:
            A Portex type instance created from the input JSON string.

        """
        return cls.from_pyobj(json.loads(content))

    @classmethod
    def from_yaml(cls: Type[_T], content: str) -> _T:
        """Create Portex type instance from YAML string.

        Arguments:
            content: A YAML string representing a Portex type.

        Returns:
            A Portex type instance created from the input YAML string.

        """
        return cls.from_pyobj(yaml.load(content, yaml.Loader))

    def to_pyobj(self, _with_imports: bool = True) -> Dict[str, Any]:
        """Dump the instance to a python dict.

        Returns:
            A python dict representation of the Portex type.

        """
        pydict: Dict[str, Any] = {}
        if _with_imports:
            imports_pyobj = self.imports.to_pyobj()
            if imports_pyobj:
                pydict["imports"] = imports_pyobj

        pydict["type"] = self.__class__.__name__
        for name, parameter in self.params.items():
            attr = getattr(self, name)
            if attr != parameter.default:
                pydict[name] = parameter.dump(attr)

        return pydict

    def to_json(self) -> str:
        """Dump the instance to a JSON string.

        Returns:
            A JSON representation of the Portex type.

        """
        return json.dumps(self.to_pyobj(), ensure_ascii=False)

    def to_yaml(self) -> str:
        """Dump the instance to a YAML string.

        Returns:
            A YAML representation of the Portex type.

        """
        return yaml.dump(self.to_pyobj(), sort_keys=False, allow_unicode=True)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        Return:
            The corresponding builtin PyArrow DataType.

        """
        raise NotImplementedError

    def to_builtin(self) -> "PortexBuiltinType":
        """Expand the top level type to Portex builtin type.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def copy(self: _T) -> _T:
        """Get a copy of the portex type.

        Returns:
            A copy of the portex type.

        """
        return deepcopy(self)


class PortexRecordBase(
    PortexType, UserMutableMapping[str, PortexType]
):  # pylint: disable=abstract-method
    """The base class of record like Portex types."""

    _fields_factory: "ConnectedFieldsFactory"

    @property
    def _data(self) -> "ConnectedFields":  # type: ignore[override]
        return self._fields_factory({name: getattr(self, name) for name in self.params})

    def _get_column_count(self) -> int:
        """Get the total column count of the record base type.

        Returns:
            The total column count.

        """
        return sum(
            portex_type._get_column_count()  # pylint: disable=protected-access
            for portex_type in self._data.values()
        )

    def insert(self, index: int, name: str, portex_type: PortexType) -> None:
        """Insert the name and portex_type at the index.

        Arguments:
            index: The index to insert the field.
            name: The name of the field to be inserted.
            portex_type: The portex_type of the field to be inserted.

        """
        self._data.insert(index, name, portex_type)

    def astype(self, name: str, portex_type: PortexType) -> None:
        """Convert the type of the field with the given name to the new PortexType.

        Arguments:
            name: The name of the field to convert.
            portex_type: The new PortexType of the field to convert to.

        """
        self._data.astype(name, portex_type)

    def rename(self, old_name: str, new_name: str) -> None:
        """Rename the name of a field.

        Arguments:
            old_name: The current name of the field to be renamed.
            new_name: The new name of the field to assign.

        """
        self._data.rename(old_name, new_name)

    def to_pyarrow(self) -> pa.StructType:
        """Convert the Portex type to the corresponding builtin PyArrow StructType.

        Returns:
            The corresponding builtin PyArrow StructType.

        """
        return pa.struct(
            [pa.field(name, portex_type.to_pyarrow()) for name, portex_type in self._data.items()]
        )


def read_yaml(path: PathLike) -> PortexType:
    """Read a yaml file into Portex type.

    Arguments:
        path: The path of the yaml file.

    Returns:
        A Portex type instance created from the input yaml file.

    """
    with open(path, encoding="utf-8") as fp:
        return PortexType.from_pyobj(yaml.load(fp, yaml.Loader))


def read_json(path: PathLike) -> PortexType:
    """Read a json file into Portex type.

    Arguments:
        path: The path of the json file.

    Returns:
        A Portex type instance created from the input json file.

    """
    with open(path, encoding="utf-8") as fp:
        return PortexType.from_pyobj(json.load(fp))



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Schema module."""

from graviti.portex.avro import convert_portex_schema_to_avro
from graviti.portex.base import PortexRecordBase, PortexType, read_json, read_yaml
from graviti.portex.builder import build_package
from graviti.portex.builtin import (
    array,
    binary,
    boolean,
    date,
    enum,
    float32,
    float64,
    int32,
    int64,
    record,
    string,
    time,
    timedelta,
    timestamp,
)
from graviti.portex.register import (
    STANDARD_URL,
    ContainerRegister,
    ExternalContainerRegister,
    ExternalElementResgister,
)

__all__ = [
    "ContainerRegister",
    "ExternalContainerRegister",
    "ExternalElementResgister",
    "PortexRecordBase",
    "PortexType",
    "STANDARD_URL",
    "array",
    "binary",
    "boolean",
    "build_package",
    "convert_portex_schema_to_avro",
    "date",
    "enum",
    "float32",
    "float64",
    "int32",
    "int64",
    "read_json",
    "read_yaml",
    "record",
    "string",
    "time",
    "timedelta",
    "timestamp",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Common tools."""

from collections import defaultdict
from datetime import datetime, timezone
from functools import wraps
from sys import version_info
from threading import Lock
from typing import Any, Callable, DefaultDict, TypeVar

_CallableWithoutReturnValue = TypeVar("_CallableWithoutReturnValue", bound=Callable[..., None])
_WEEKS = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
_MONTHS = ("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")

locks: DefaultDict[int, Lock] = defaultdict(Lock)


def urlnorm(url: str) -> str:
    """Normalized the input url by removing the trailing slash.

    Arguments:
        url: the url needs to be normalized.

    Returns:
        The normalized url.

    """
    return url.rstrip("/")


def locked(func: _CallableWithoutReturnValue) -> _CallableWithoutReturnValue:
    """The decorator to add threading lock for methods.

    Arguments:
        func: The method needs to add threading lock.

    Returns:
        The method with theading locked.

    """

    @wraps(func)
    def wrapper(self: Any, *arg: Any, **kwargs: Any) -> None:
        key = id(self)
        lock = locks[key]
        acquire = lock.acquire(blocking=False)
        try:
            if acquire:
                func(self, *arg, **kwargs)
                del locks[key]
            else:
                lock.acquire()
        finally:
            lock.release()

    return wrapper  # type: ignore[return-value]


def shorten(origin: str) -> str:
    """Return the first 7 characters of the original string.

    Arguments:
        origin: The string needed to be shortened.

    Returns:
        A string of length 7.

    """
    return origin[:7]


if version_info >= (3, 7):

    def convert_iso_to_datetime(date_string: str) -> datetime:
        """Convert iso 8601 format string to datetime format time with local timezone.

        Arguments:
            date_string: The iso 8601 format string.

        Returns:
            The datetime format time with local timezone.

        """
        return datetime.fromisoformat(date_string.replace("Z", "+00:00")).astimezone()

else:
    _DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

    def convert_iso_to_datetime(date_string: str) -> datetime:
        """Convert iso 8601 format string to datetime format time with local timezone.

        Arguments:
            date_string: The iso 8601 format string.

        Returns:
            The datetime format time with local timezone.

        """
        return (
            datetime.strptime(date_string, _DATE_FORMAT).replace(tzinfo=timezone.utc).astimezone()
        )


def convert_datetime_to_gmt(utctime: datetime) -> str:
    """Convert datetime to gmt format string.

    Arguments:
        utctime: The datetime with utc timezone.

    Returns:
        The gmt format string.

    """
    return (
        f"{_WEEKS[utctime.weekday()]}, {utctime.day:02d} {_MONTHS[utctime.month - 1]}"
        f" {utctime.year:04d} {utctime.hour:02d}:{utctime.minute:02d}:{utctime.second:02d} GMT"
    )


class ModuleMocker:
    """A fake module to raise ``ModuleNotFoundError`` lazily.

    Arguments:
        message: The error message for the raised ``ModuleNotFoundError``.

    """

    def __init__(self, message: str) -> None:
        self._message = message

    def __call__(self, *_: Any, **__: Any) -> Any:
        """Raise ``ModuleNotFoundError`` when called.

        Arguments:
            _: Useless positional arguments.
            __: Useless keyword arguments

        Raises:
            ModuleNotFoundError: When called.

        """
        raise ModuleNotFoundError(super().__getattribute__("_message"))

    def __getattribute__(self, name: str) -> Any:
        raise ModuleNotFoundError(super().__getattribute__("_message"))
